
# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Type
from typing import Union

import copy
import traceback

from logging import getLogger
from logging import Logger
from inspect import iscoroutinefunction

from langchain_core.messages.base import BaseMessage

from leaf_common.config.resolver_util import ResolverUtil

from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.chat.chat_history_message_processor import ChatHistoryMessageProcessor
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.internals.graph.activations.sly_data_redactor import SlyDataRedactor
from neuro_san.internals.interfaces.front_man import FrontMan
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_framework_message import AgentFrameworkMessage
from neuro_san.internals.messages.base_message_dictionary_converter import BaseMessageDictionaryConverter
from neuro_san.internals.run_context.factory.run_context_factory import RunContextFactory
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.message_processing.message_processor import MessageProcessor
from neuro_san.message_processing.answer_message_processor import AnswerMessageProcessor
from neuro_san.message_processing.structure_message_processor import StructureMessageProcessor

# Lazily import specific errors from llm providers
PATIENCE_ERRORS: Tuple[Type[Any], ...] = ResolverUtil.create_type_tuple([
                                            "openai.BadRequestError",
                                         ])


# pylint: disable=too-many-instance-attributes
class DataDrivenChatSession:
    """
    ChatSession implementation that consolidates policy
    in using data-driven agent tool graphs.
    """

    def __init__(self, agent_network: AgentNetwork):
        """
        Constructor

        :param agent_network: The AgentNetwork to use.
        """
        # We make a copy of the AgentNetwork at this level to allow for interactions
        # within this scope to modify the network topology.
        agent_network_copy: AgentNetwork = copy.deepcopy(agent_network)
        self.registry: AgentToolRegistry = AgentToolRegistry(agent_network_copy)

        self.front_man: FrontMan = None
        self.sly_data: Dict[str, Any] = {}

    async def set_up(self, invocation_context: InvocationContext,
                     chat_context: Dict[str, Any] = None):
        """
        Resets or sets the instance up for the first time.
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        """

        # Reset any sly data
        # This ends up being the one reference to the sly_data that gets passed around
        # to the graph components. Updating this updates everyone else.
        self.sly_data = {}

        # Reset what we might have created before.
        await self.delete_resources()

        run_context: RunContext = RunContextFactory.create_run_context(None, None,
                                                                       invocation_context=invocation_context,
                                                                       chat_context=chat_context)

        self.front_man = self.registry.create_front_man(self.sly_data, run_context)

        await self.front_man.create_any_resources()

    async def chat(self, user_input: str,
                   invocation_context: InvocationContext,
                   sly_data: Dict[str, Any] = None) -> Iterator[Dict[str, Any]]:
        """
        Main entry-point method for accepting new user input

        :param user_input: A string with the user's input
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be None.
        :return: An Iterator over dictionary representation of chat messages.
                The keys/values/structure of these chat message dictionaries will reflect
                instances of ChatMessage from chat.proto.

                Note that Iterators themselves are *not* simply lists. They are a Python
                construct intended for use in a for-loop that is allowed to come up with
                its content dynamically.  For our purposes, when an initiator of chat()
                gets a handle to this Iterator, they can begin looping/waiting on its contents
                without the content itself having been created yet.  This is a building
                block of streaming results even though direct callers may not actually
                be streaming.
        """
        if self.front_man is None:
            await self.set_up(invocation_context)
        else:
            self.front_man.update_invocation_context(invocation_context)

        # Update sly data, if any.
        # Note that since this instance is the owner of the sly_data,
        # any update here should get transmitted to all the other graph components
        # because it is expected they share the reference and only interact with it
        # in a read-only fashion.
        if sly_data is not None:
            self.sly_data.update(sly_data)

        try:
            # DEF - drill further down for iterator from here to enable getting
            #       messages from downstream agents.
            raw_messages: List[BaseMessage] = await self.front_man.submit_message(user_input)

        except PATIENCE_ERRORS:
            # This can happen if the user is trying to send a new message
            # while it is still working on a previous message that has not
            # yet returned.
            raw_messages: List[BaseMessage] = [
                AgentFrameworkMessage(content="Patience, please. I'm working on it.")
            ]

            logger: Logger = getLogger(self.__class__.__name__)
            logger.error(traceback.format_exc())

        converter = BaseMessageDictionaryConverter(origin=self.front_man.get_origin())
        chat_messages: List[Dict[str, Any]] = []
        for raw_message in raw_messages:
            chat_message: Dict[str, Any] = converter.to_dict(raw_message)
            chat_messages.append(chat_message)

        return iter(chat_messages)

    # pylint: disable=too-many-locals
    async def streaming_chat(self, user_input: str,
                             invocation_context: InvocationContext,
                             sly_data: Dict[str, Any] = None,
                             chat_context: Dict[str, Any] = None):
        """
        Main streaming entry-point method for accepting new user input

        :param user_input: A string with the user's input
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be None.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        :return: Nothing.  Response values are put on a queue whose consumtion is
                managed by the Iterator aspect of AsyncCollatingQueue on the InvocationContext.
        """
        if self.front_man is None:
            await self.set_up(invocation_context, chat_context)

        # Save information about chat
        chat_messages: Iterator[Dict[str, Any]] = await self.chat(user_input, invocation_context, sly_data)
        message_list: List[Dict[str, Any]] = list(chat_messages)

        # Determine the chat_context to enable continuing the conversation
        return_chat_context: Dict[str, Any] = self.prepare_chat_context(message_list)

        # Get the front man spec. We will need it later for a few things.
        front_man_spec: Dict[str, Any] = self.front_man.get_agent_tool_spec()

        # Get the formats we should parse from the final answer from the config for the network.
        # As of 6/24/25, this is an unadvertised experimental feature.
        structure_formats: Union[str, List[str]] = front_man_spec.get("structure_formats")

        # Find "the answer" and have that be the content of the last message we send
        answer_processor = AnswerMessageProcessor(structure_formats=structure_formats)
        answer_processor.process_messages(message_list)
        answer: str = answer_processor.get_answer()
        if answer is None:
            # Can't have content as None or problems arise.
            answer = ""
        structure: Dict[str, Any] = answer_processor.get_structure()

        # Send back sly_data as the front-man permits
        redactor = SlyDataRedactor(front_man_spec,
                                   config_keys=["allow.to_upstream.sly_data"],
                                   allow_empty_dict=False)
        return_sly_data: Dict[str, Any] = redactor.filter_config(self.sly_data)

        # Stream over chat state as the last message
        message = AgentFrameworkMessage(content=answer, chat_context=return_chat_context,
                                        sly_data=return_sly_data, structure=structure)
        journal: Journal = invocation_context.get_journal()
        await journal.write_message(message, origin=None)

        # Put an end-marker on the queue to tell the consumer we truly are done
        # and it doesn't need to wait for any more messages.
        # The consumer await-s for queue.get()
        queue: AsyncCollatingQueue = invocation_context.get_queue()

        # The synchronous=True is necessary when an async HTTP request is at the get()-ing end of the queue,
        # as the journal messages come from inside a separate event loop from that request. The lock
        # taken here ends up being harmless in the synchronous request case (like for gRPC) because
        # we would only be blocking our own event loop.
        await queue.put_final_item(synchronous=True)

        # Now that we are done, tell the Reservationist that we used for this request
        # that there will be no more Reservations to corral.
        reservationist: Reservationist = invocation_context.get_reservationist()
        if reservationist is not None:
            await reservationist.close()

        # Close any objects on sly data that can be closed.
        await self.close_sly_data()

    async def delete_resources(self):
        """
        Frees up any service-side resources.
        """
        if self.front_man is not None:
            await self.front_man.delete_any_resources()
            self.front_man = None

    def prepare_chat_context(self, chat_message_history: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Prepare the chat context.

        :param chat_message_history: A list of ChatMessage dictionaries that
                comprise the front man's full chat history.
        :return: A ChatContext dictionary comprising the full state of play of
                the conversation such that it could be taken up on a different
                server instance
        """
        # OK if this is None
        max_message_history: int = self.front_man.get_agent_tool_spec().get("max_message_history")
        processor: MessageProcessor = ChatHistoryMessageProcessor(max_message_history)
        processor.process_messages(chat_message_history)

        chat_history: Dict[str, Any] = {
            "origin": self.front_man.get_origin(),
            "messages": processor.get_message_history()
        }

        # For now, we only send the front man's chat history, as that is the
        # state we had been preserving since the early days.  It is conceivable
        # this could expand to other agents in the hierarchy at some point.
        chat_context: Dict[str, Any] = {
            "chat_histories": [chat_history]
        }

        return chat_context

    def create_outgoing_message_processor(self) -> MessageProcessor:
        """
        :return: A MessageProcessor that filters messages outgoing to the client.
                How this works is based on settings on the front man.
                Can be None.
        """
        message_processor: MessageProcessor = None

        front_man_name: str = self.registry.find_front_man()
        front_man_spec: Dict[str, Any] = self.registry.get_agent_tool_spec(front_man_name)

        # Get the formats we should parse from the final answer from the config for the network.
        # As of 6/24/25, this is an unadvertised experimental feature.
        structure_formats: Union[str, List[str]] = front_man_spec.get("structure_formats")
        if structure_formats is None:
            return message_processor

        # Eventually this might be a CompositeMessageProcessor
        message_processor = StructureMessageProcessor(structure_formats)
        return message_processor

    async def close_sly_data(self):
        """
        Close any sly data value that can be closed.
        """
        if self.sly_data is None or len(self.sly_data) == 0:
            # Nothing to close
            return

        # Try closing anything that can be close()-d in the sly_data.
        for key, value in self.sly_data.items():
            # Note: It's possible we might want to try closing keys as well
            description: str = f"sly_data[{key}]"
            await self.close_one(value, description)

    async def close_one(self, value: Any, description: str):
        """
        Close a single object if it can be closed.
        :param value: The object to close
        :param description: Where the object to close came from
        """
        if not hasattr(value, "close"):
            # Nothing to close
            return

        close_method = getattr(value, "close")
        if not callable(close_method):
            # No method to call
            return

        # Call the close method
        try:
            if iscoroutinefunction(close_method):
                await close_method()
            else:
                close_method()

        except TypeError:
            # Not enough arguments to call close().
            logger: Logger = getLogger(self.__class__.__name__)
            logger.error("Unable to close() %s", description)
            logger.error(traceback.format_exc())




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from logging import getLogger
from logging import Logger

from leaf_common.config.config_filter import ConfigFilter


class ManifestKeyConfigFilter(ConfigFilter):
    """
    Implementation of the ConfigFilter interface that reads the contents
    of a single manifest file for agent networks/registries, converting
    keys to a standardized form.
    """

    def __init__(self, manifest_file: str):
        """
        Constructor

        :param manifest_file: The name of the manifest file we are processing for logging purposes
        """
        self.logger: Logger = getLogger(self.__class__.__name__)
        self.manifest_file: str = manifest_file

    def filter_config(self, basis_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Filters the given basis config.

        :param basis_config: The config dictionary to act as the basis
                for filtering
        :return: A config dictionary, potentially modified as per the
                policy encapsulated by the implementation
        """

        filtered: Dict[str, Any] = {}

        for key, value in basis_config.items():

            # Key here is an agent name in a form that we choose.
            # Keys sometimes come with quotes.
            manifest_key: str = key.replace(r'"', "")
            manifest_key = manifest_key.strip()

            filtered[manifest_key] = value

        return filtered




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from copy import copy

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.journals.originating_journal import OriginatingJournal
from neuro_san.internals.messages.base_message_dictionary_converter import BaseMessageDictionaryConverter
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor


class ExternalMessageProcessor(MessageProcessor):
    """
    MessageProcessor implementation for handling messages from an external agent network.
    """

    def __init__(self, journal: OriginatingJournal):
        """
        Constructor

        :param journal: The OriginatingJournal through which messages from
                    the external agent are passed.
        """
        self.journal: OriginatingJournal = journal

    async def async_process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message asynchronously.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        message_origin: List[Dict[str, Any]] = chat_message_dict.get("origin")
        if message_origin is None:
            return

        # Append the origin information from the external agent to our own
        origin: List[Dict[str, Any]] = copy(self.journal.get_origin())
        origin.extend(message_origin)

        # Send the message to the client with deepened origin information
        converter = BaseMessageDictionaryConverter(langchain_only=False)
        message: BaseMessage = converter.from_dict(chat_message_dict)
        await self.journal.write_message(message, origin=origin)

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        # We don't implement this because we need to do so asynchronously
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List


class DictionaryValidator:
    """
    An interface for validating dictionaries of various meanings.
    """

    def validate(self, candidate: Dict[str, Any]) -> List[str]:
        """
        Validate the given dictionary

        :param candidate: The dictionary to validate
        :return: A list of error messages
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Dict

from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider


class SingleAgentNetworkProvider(AgentNetworkProvider):
    """
    Class providing current AgentNetwork for a given agent in the service scope.
    """
    def __init__(self, agent_name: str, agents_table: Dict[str, AgentNetwork]):
        """
        Constructor.
        :param agent_name: name of an agent to provide AgentNetwork instances for;
        :param agents_table: service-wide table mapping agent names to their
            currently active AgentNetwork instances.
            This table is assumed to be dynamically modified outside a single agent scope.
        """
        self.agent_name = agent_name
        self.agents_table: Dict[str, AgentNetwork] = agents_table

    def get_agent_network(self) -> AgentNetwork:
        """
        :return: Current AgentNetwork instance for specific agent name.
                None if this does not exist for the instance's agent_name.
        """
        return self.agents_table.get(self.agent_name)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any

from neuro_san.message_processing.basic_message_processor import BasicMessageProcessor


class AgentEvaluator:
    """
    Interface definition for evaluating part of an agent's response
    """

    def evaluate(self, processor: BasicMessageProcessor, test_key: str, verify_for: Any):
        """
        Evaluate the contents of the BasicMessageProcessor

        :param processor: The BasicMessageProcessor to evaluate
        :param test_key: the compound .-delimited key of the response value to test
        :param verify_for: The data to evaluate the response against
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Tuple

import os
import time

from json import dumps
from json import loads
from json.decoder import JSONDecodeError
from logging import getLogger
from logging import Logger

from boto3 import client as boto3_client
from botocore.client import BaseClient
from botocore.exceptions import ClientError
from botocore.exceptions import NoCredentialsError

from neuro_san.interfaces.reservation import Reservation
from neuro_san.internals.interfaces.reservations_storage import ReservationsStorage
from neuro_san.service.watcher.interfaces.startable import Startable
from neuro_san.service.watcher.temp_networks.reservation_dictionary_converter import ReservationDictionaryConverter


class S3ReservationsStorage(ReservationsStorage, Startable):
    """
    AWS S3-based implementation of ReservationsStorage.

    Stores reservations as JSON objects in an S3 bucket, with each reservation
    stored in its associated agent spec as metadata.
    """

    def __init__(self, bucket_name: str = "", prefix: str = "reservations/"):
        """
        Initialize S3 reservations storage.

        :param bucket_name: S3 bucket name (defaults to AGENT_RESERVATIONS_S3_BUCKET env var)
        :param prefix: S3 key prefix for reservation objects
        """
        # Configure bucket name from parameter or environment variable
        env_bucket: str = os.getenv("AGENT_RESERVATIONS_S3_BUCKET", "")
        self.bucket_name: str = bucket_name or env_bucket
        if not self.bucket_name:
            raise ValueError(
                "S3 bucket name must be provided via bucket_name parameter or "
                "AGENT_RESERVATIONS_S3_BUCKET environment variable"
            )

        # Set up S3 key prefix and initialize sync target
        self.prefix: str = prefix
        self.sync_target: ReservationsStorage = None
        self.s3_client: BaseClient = None

        # Track last sync timestamp for incremental syncing (0.0 means sync all)
        self.last_sync_timestamp: float = 0.0
        self.converter = ReservationDictionaryConverter()

        self.logger: Logger = getLogger(self.__class__.__name__)

    def start(self):
        """
        Initialize the S3 client and validate connection to the bucket.

        This method can be called to re-initialize the connection if needed.
        """
        try:
            # Initialize S3 client using default AWS credential chain
            self.s3_client = boto3_client("s3")

            # Validate bucket exists and we have access by performing a head operation
            self.s3_client.head_bucket(Bucket=self.bucket_name)
            self.logger.info("Successfully connected to S3 bucket: %s", self.bucket_name)

        except NoCredentialsError as exception:
            # Handle missing AWS credentials
            raise ValueError("AWS credentials not found. Please configure AWS credentials.") from exception
        except ClientError as exception:
            # Handle various S3 access errors with specific messages
            error_code: str = exception.response["Error"]["Code"]
            if error_code == "404":
                raise ValueError(f"S3 bucket '{self.bucket_name}' does not exist") from exception
            if error_code == "403":
                raise ValueError(f"Access denied to S3 bucket '{self.bucket_name}'") from exception
            raise ValueError(f"Error accessing S3 bucket '{self.bucket_name}': {exception}") from exception

    def set_sync_target(self, sync_target: ReservationsStorage):
        """
        Set the sync target for this storage implementation.

        :param sync_target: The ReservationsStorage where in-memory versions end up
        """
        # Store reference to the target storage for syncing operations
        # This allows S3 storage to push reservations to in-memory or other storage types
        self.sync_target = sync_target

    def add_reservations(self, reservations_dict: Dict[Reservation, Any],
                         source: str = None):
        """
        Add reservations to S3 storage.

        :param reservations_dict: A mapping of Reservation -> some deployable agent spec
        :param source: A string describing where the deployment was coming from
        """
        self.logger.info("Adding %d reservations to S3", len(reservations_dict))

        # Process each reservation/agent spec pair individually
        reservation: Reservation = None
        agent_spec: Dict[str, Any] = None
        for reservation, agent_spec in reservations_dict.items():

            # Build complete data structure containing reservation metadata,
            # the associated agent_spec, source information, and storage timestamp
            current_time: float = time.time()
            agent_spec["metadata"] = {
                "reservation": self.converter.to_dict(reservation),  # Serialized reservation object
                "stored_at": current_time              # When stored in S3
            }

            # Generate S3 key using prefix and reservation ID for easy lookup
            reservation_id: str = reservation.get_reservation_id()
            key: str = f"{self.prefix}{reservation_id}.json"

            # Store as JSON object in S3 with proper content type
            json_body: str = dumps(agent_spec, indent=4)  # Pretty-printed JSON
            self.s3_client.put_object(
                Bucket=self.bucket_name,
                Key=key,
                Body=json_body,
                ContentType="application/json"
            )

            self.logger.debug("Successfully stored reservation %s in S3", reservation_id)

    def sync_one_reservation(self, obj_key: str) -> Tuple[Reservation, Any]:
        """
        Sync a single reservation from S3.

        :param obj_key: S3 object key for the reservation
        :return: Tuple of (reservation, agent_spec) if successful and not expired,
                 (None, None) otherwise
        """
        reservation: Reservation = None
        agent_spec: Dict[str, Any] = None
        try:
            # Retrieve the reservation object from S3
            obj_response: Dict[str, Any] = self.s3_client.get_object(
                Bucket=self.bucket_name,
                Key=obj_key
            )

            # Parse JSON content from S3 object body
            json_content: str = obj_response["Body"].read().decode("utf-8")
            agent_spec: Dict[str, Any] = loads(json_content)
            metadata: Dict[str, Any] = agent_spec.get("metadata")

            # Reconstruct the Reservation object from stored dictionary
            reservation_dict: Dict[str, Any] = metadata.get("reservation")
            reservation = self.converter.from_dict(reservation_dict)

            self.logger.debug("Successfully synced active reservation %s",
                              reservation.get_reservation_id())

        except ClientError as exception:
            # Handle case where another process already removed the object before we could read it
            if exception.response["Error"]["Code"] == "NoSuchKey":
                self.logger.debug("Reservation %s was already removed by another process during sync",
                                  obj_key)
            else:
                # Log other S3 errors but don't raise - allows sync to continue
                self.logger.error("S3 error processing reservation object %s during sync: %s",
                                  obj_key, str(exception))

        except JSONDecodeError as exception:
            # Log JSON errors but don't raise - allows sync to continue
            self.logger.error("JSON error processing reservation object %s during sync: %s",
                              obj_key, str(exception))

        return reservation, agent_spec

    def sync_reservations(self):
        """
        Sync reservations from S3 to the sync target (if set).
        Only syncs objects modified after the last sync timestamp for efficiency.
        """
        # Validate that a sync target has been configured
        if not self.sync_target:
            self.logger.warning("No sync target set, skipping sync operation")
            return

        sync_start_time: float = time.time()

        if self.last_sync_timestamp > 0.0:
            self.logger.debug("Starting incremental sync from S3 (since %s)",
                              time.ctime(self.last_sync_timestamp))
        else:
            self.logger.debug("Starting full sync operation from S3 to configured sync target")

        # List all reservation objects in S3 bucket with our prefix
        response: Dict[str, Any] = self.s3_client.list_objects_v2(
            Bucket=self.bucket_name,
            Prefix=self.prefix
        )

        # Handle case where no reservations exist in S3
        if "Contents" not in response:
            self.logger.debug("No reservations found in S3 bucket")
            # Update sync timestamp even if no objects found
            self.last_sync_timestamp = sync_start_time
            return

        # Build dictionary of active reservations to sync
        reservations_dict: Dict[Reservation, Any] = {}
        processed_count: int = 0
        skipped_count: int = 0

        # Process each S3 object individually using our helper method
        obj: Dict[str, Any]
        for obj in response["Contents"]:
            # Skip objects that haven't been modified since last sync
            if self.last_sync_timestamp > 0.0:
                obj_modified_time: float = obj["LastModified"].timestamp()
                if obj_modified_time <= self.last_sync_timestamp:
                    skipped_count += 1
                    continue

            processed_count += 1
            reservation: Reservation = None
            agent_spec: Dict[str, Any] = None
            reservation, agent_spec = self.sync_one_reservation(obj["Key"])
            if reservation is None or agent_spec is None:
                # Skip anything that had an error associated with it
                continue

            # Only include successfully processed, non-expired reservations (skip empty ones)
            if reservation.get_reservation_id() != "" and agent_spec != {}:
                reservations_dict[reservation] = agent_spec

        # Push collected reservations to the sync target storage
        if reservations_dict:
            self.sync_target.add_reservations(reservations_dict)
            self.logger.info("Successfully synced %d active reservations to target " +
                             "(processed %d, skipped %d unchanged objects)",
                             len(reservations_dict), processed_count, skipped_count)
        else:
            if processed_count > 0:
                self.logger.debug(
                    "No active reservations found to sync (processed %d, skipped %d unchanged objects)",
                    processed_count, skipped_count)
            else:
                self.logger.debug("No new objects to sync (skipped %d unchanged objects)", skipped_count)

        # Update the last sync timestamp to mark successful completion
        self.last_sync_timestamp = sync_start_time

    def expire_one_reservation(self, obj_key: str, current_time: float) -> bool:
        """
        Check and expire a single reservation if it's expired.

        :param obj_key: S3 object key for the reservation
        :param current_time: Current timestamp to compare against
        :return: True if reservation was expired and deleted, False otherwise
        """
        expired: bool = False
        try:
            # Retrieve the reservation object from S3
            obj_response: Dict[str, Any] = self.s3_client.get_object(
                Bucket=self.bucket_name,
                Key=obj_key
            )

            # Parse JSON content to extract reservation metadata
            json_content: str = obj_response["Body"].read().decode("utf-8")
            agent_spec: Dict[str, Any] = loads(json_content)
            metadata: Dict[str, Any] = agent_spec.get("metadata")
            reservation_data: Dict[str, Any] = metadata.get("reservation")

            # Compare current time against reservation's expiration timestamp
            expiration_time: float = reservation_data.get("expiration_time_in_seconds")
            if current_time > expiration_time:
                # Reservation has expired - remove it from S3 storage
                try:
                    self.s3_client.delete_object(
                        Bucket=self.bucket_name,
                        Key=obj_key
                    )
                    reservation_id: str = reservation_data.get("id")
                    self.logger.debug("Deleted expired reservation %s from S3", reservation_id)
                    expired = True
                except ClientError as delete_error:
                    # Handle case where another process already deleted the object
                    if delete_error.response["Error"]["Code"] != "NoSuchKey":
                        # Re-raise other delete errors
                        raise delete_error

                    self.logger.debug("Reservation %s was already deleted by another process", obj_key)
                    expired = True  # Consider this a successful expiration

            # Reservation is still active - no action needed

        except ClientError as exception:
            # Handle case where another process already removed the object before we could read it
            if exception.response["Error"]["Code"] == "NoSuchKey":
                self.logger.debug("Reservation %s was already removed by another process", obj_key)
                expired = True  # Object is gone, which is the desired outcome for expiration
            else:
                # Log other S3 errors but don't raise - allows expiration to continue
                self.logger.error("S3 error processing reservation object %s: %s", obj_key, str(exception))

        except JSONDecodeError as exception:
            # Log JSON errors but don't raise - allows expire to continue
            self.logger.error("JSON error processing reservation object %s during expire: %s",
                              obj_key, str(exception))

        return expired

    def expire_reservations(self):
        """
        Remove expired reservations from S3 storage.
        """
        self.logger.debug("Starting expiration process for S3 reservations")

        # List all reservation objects in S3 bucket with our prefix
        response: Dict[str, Any] = self.s3_client.list_objects_v2(
            Bucket=self.bucket_name,
            Prefix=self.prefix
        )

        # Handle case where no reservations exist in S3
        if "Contents" not in response:
            self.logger.debug("No reservations found in S3 bucket for expiration")
            return

        # Track how many reservations we expire for reporting
        expired_count: int = 0
        # Get current timestamp once for consistent expiration checking
        current_time: float = time.time()

        # Process each S3 object individually using our helper method
        obj: Dict[str, Any]
        for obj in response["Contents"]:
            # Attempt to expire this reservation and increment counter if successful
            if self.expire_one_reservation(obj["Key"], current_time):
                expired_count += 1

        if expired_count > 0:
            self.logger.info("Expiration complete: removed %d expired reservations from S3", expired_count)
        else:
            self.logger.debug("Expiration complete: removed no expired reservations from S3")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import List

from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator
from neuro_san.internals.validation.common.composite_dictionary_validator import CompositeDictionaryValidator
from neuro_san.internals.validation.network.cycles_network_validator import CyclesNetworkValidator
from neuro_san.internals.validation.network.missing_nodes_network_validator import MissingNodesNetworkValidator
from neuro_san.internals.validation.network.unreachable_nodes_network_validator import UnreachableNodesNetworkValidator


class StructureNetworkValidator(CompositeDictionaryValidator):
    """
    Implementation of CompositeDictionaryValidator interface that uses multiple specific validators
    to do some standard validation for topological issues.
    This gets used by agent network designer.
    """

    def __init__(self):
        """
        Constructor
        """
        validators: List[DictionaryValidator] = [
            CyclesNetworkValidator(),
            MissingNodesNetworkValidator(),
            UnreachableNodesNetworkValidator(),
        ]
        super().__init__(validators)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Dict

import logging

from neuro_san.interfaces.async_agent_session import AsyncAgentSession
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.async_agent_session_factory import AsyncAgentSessionFactory
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.session.async_direct_agent_session import AsyncDirectAgentSession
from neuro_san.session.async_http_service_agent_session import AsyncHttpServiceAgentSession


class ExternalAgentSessionFactory(AsyncAgentSessionFactory):
    """
    Creates AgentSessions for external agents.
    """

    def __init__(self, use_direct: bool = False,
                 network_storage_dict: Dict[str, AgentNetworkStorage] = None):
        """
        Constructor

        :param use_direct: When True, will use a Direct session for
                    external agents that would reside on the same server.
        :param network_storage_dict: A dictionary of AgentNetworkStorage instances which keeps all
                                the AgentNetwork instances.  Only used with use_direct=True.
        """
        self.use_direct: bool = use_direct
        self.network_storage_dict: Dict[str, AgentNetworkStorage] = None
        if self.use_direct:
            # Only store the reference if we will need it
            self.network_storage_dict = network_storage_dict

    def create_session(self, agent_url: str,
                       invocation_context: InvocationContext) -> AsyncAgentSession:
        """
        :param agent_url: An url string pointing to an external agent that came from
                    a tools list in an agent spec.
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :return: An AsyncAgentSession through which communications about the external agent can be made.
        """

        agent_location: Dict[str, str] = ExternalAgentParsing.parse_external_agent(agent_url)
        session: AsyncAgentSession = self.create_session_from_location_dict(agent_location, invocation_context)
        return session

    def create_session_from_location_dict(self, agent_location: Dict[str, str],
                                          invocation_context: InvocationContext) -> AsyncAgentSession:
        """
        :param agent_location: An agent location dictionary returned by
                    ExternalAgentParsing.parse_external_agent()
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :return: An AsyncAgentSession through which communications about the external agent can be made.
        """
        if agent_location is None:
            return None
        # Create the session.
        host = agent_location.get("host")
        port = agent_location.get("port")
        agent_name = agent_location.get("agent_name")

        # Note: It's possible we might want some filtering/translation of
        #       metadata keys not unlike what we are doing for sly_data.
        metadata: Dict[str, str] = None
        if invocation_context is not None:
            metadata = invocation_context.get_metadata()

        session: AsyncAgentSession = None
        if self.use_direct and (host is None or len(host) == 0 or host == "localhost"):

            # Optimization: We want to create a different kind of session to minimize socket usage
            # and potentially relieve the direct user of the burden of having to start a server

            agent_network_provider: AgentNetworkProvider = None
            for network_storage in self.network_storage_dict.values():
                # Be sure we have something
                agent_network_provider = network_storage.get_agent_network_provider(agent_name)
                if agent_network_provider.get_agent_network() is not None:
                    break

            agent_network: AgentNetwork = agent_network_provider.get_agent_network()
            safe_invocation_context: InvocationContext = invocation_context.safe_shallow_copy()
            session = AsyncDirectAgentSession(agent_network, safe_invocation_context, metadata=metadata)

        if session is None:
            # When creating a session for external agents, specifically use None for the
            # streaming timeout.  This implies an infinite amount of time to let the
            # external agent get its job done.  The rationale here is that:
            #   a)  We do not know how long any given external agent is really going to take
            #       to do its job.
            #   b)  We figure that the regular connection aspects to the server in question
            #       have already been sorted out in the obligitory call to function() that
            #       precedes any streaming_chat() call.
            session = AsyncHttpServiceAgentSession(host, port, agent_name=agent_name,
                                                   metadata=metadata, streaming_timeout_in_seconds=None)

        # Quiet any logging from leaf-common grpc stuff.
        quiet_please = logging.getLogger("leaf_common.session.grpc_client_retry")
        quiet_please.setLevel(logging.WARNING)

        return session

    def is_use_direct(self) -> bool:
        """
        :return: When True, will use a Direct session for external agents that would reside on the same server.
        """
        return self.use_direct




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.internals.graph.activations.abstract_class_activation import AbstractClassActivation
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext


class ToolboxActivation(AbstractClassActivation):
    """
    A ClassActivation that resolves the full class reference from a predefined coded tool in the toolbox.

    Note that this class does not apply to Langchain's base tools.
    """

    def get_full_class_ref(self) -> str:
        """
        Returns the full class reference path from a predefined toolbox.

        This implementation looks up the tool by name in a toolbox, using the
        "toolbox" field in `agent_tool_spec`, then determine the class of that
        coded tool.

        :return: A dot-separated string representing the full class path.
        """
        tool_name: str = self.agent_tool_spec.get("toolbox")
        invocation_context: InvocationContext = self.run_context.get_invocation_context()
        toolbox_factory: ContextTypeToolboxFactory = invocation_context.get_toolbox_factory()
        return toolbox_factory.get_shared_coded_tool_class(tool_name)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from logging import getLogger
from logging import Logger


class AsyncCloseable:
    """
    A simple object used to test the asynchronous close()-ing of objects on sly_data.
    """

    async def close(self):
        """
        Close the object
        """
        logger: Logger = getLogger(self.__class__.__name__)
        logger.info("async close() called")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from neuro_san.service.http.logging.log_context_filter import LogContextFilter as ParentLogContextFilter

print("""
WARNING: The class:
     neuro_san.http_sidecar.logging.log_context_filter.LogContextFilter
... has moved to be ...
    neuro_san.service.http.logging.log_context_filter.LogContextFilter.
Please update the logging.json file associated with your Neuro SAN agent server accordingly.
""")


class LogContextFilter(ParentLogContextFilter):
    """
    Compatibility class for logging.json entries that haven't quite converted yet.
    """




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Set
from typing import Type
from typing import Union
from types import MethodType
from inspect import isclass
from inspect import signature
from pydantic import BaseModel


class ArgumentValidator:
    """
    A utility class for inspecting method and class arguments, particularly useful for
    validating input against Pydantic `BaseModel` subclasses or callable method signatures.

    This class provides static methods to:
    - Validate that a dictionary of arguments matches the accepted parameters of a class or method.
    - Extract all field names and aliases from a Pydantic BaseModel subclass.
    """

    @staticmethod
    def check_invalid_args(method_class: Union[Type, MethodType], args: Dict[str, Any]):
        """
        Check for invalid arguments in a class constructor or method call.

        :param method_class: The class or method to validate against.
        :param args: Dictionary of argument to check.
        :raises ValueError: If any argument name is not accepted by the method or class.
        """

        # If method_class is a Pydantic BaseModel, get its field names and aliases
        if isclass(method_class) and issubclass(method_class, BaseModel):
            class_args_set: Set[str] = ArgumentValidator.get_base_model_args(method_class)
        else:
            # Otherwise, extract argument names from the function/method signature
            class_args_set = set(signature(method_class).parameters.keys())

        # Get the argument keys provided by the user
        args_set: Set[str] = set(args.keys())

        # Identify which arguments are not accepted by the method/class
        invalid_args: Set[str] = args_set - class_args_set
        if invalid_args:
            raise ValueError(
                f"Arguments {invalid_args} for '{method_class.__name__}' do not match any attributes "
                "of the class or any arguments of the method."
            )

    @staticmethod
    def get_base_model_args(base_model_class: Type[BaseModel]) -> Set[str]:
        """
        Extract all field names and aliases from a Pydantic BaseModel class.

        :param base_model_class: A class that inherits from `BaseModel`.
        :return: A set of valid argument names, including both field names and aliases.
        """

        fields_and_aliases: Set[str] = set()

        # Check for field name and info
        # field info includes attributes like "required", "default", "description", and "alias"
        for field_name, field_info in base_model_class.model_fields.items():
            # Add field name to the set
            fields_and_aliases.add(field_name)
            if field_info.alias:
                # If there is "alias" in the info add it to the set as well
                fields_and_aliases.add(field_info.alias)

        return fields_and_aliases




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.internals.parsers.structure.json_structure_parser import JsonStructureParser
from neuro_san.internals.parsers.structure.structure_parser import StructureParser


class StructureParserFactory:
    """
    Factory for creating StructureParser instances based on a string type
    """

    def create_structure_parser(self, parser_type: str) -> StructureParser:
        """
        Creates a structure parser given the string type

        :param parser_type: A string describing the format of the structure parser.
        """

        structure_parser: StructureParser = None

        if parser_type is None or not isinstance(parser_type, str):
            structure_parser = None
        elif parser_type.lower() == "json":
            structure_parser = JsonStructureParser()

        return structure_parser




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from os import environ

from leaf_common.config.resolver_util import ResolverUtil

from neuro_san.interfaces.usage_logger import UsageLogger
from neuro_san.service.usage.wrapped_usage_logger import WrappedUsageLogger


class UsageLoggerFactory:
    """
    Implementation of the UsageLogger interface that merely spits out
    usage stats to the logger.
    """

    @staticmethod
    def create_usage_logger() -> WrappedUsageLogger:
        """
        Reads the server environment variables to create a UsageLogger instance.

        :return: A WrappedUsageLogger that wraps the class referred to by the
                AGENT_USAGE_LOGGER env var.  Can throw an exception
                if there are problems creating the class referenced by the env var.
        """
        usage_logger_class_name: str = environ.get("AGENT_USAGE_LOGGER")
        usage_logger: UsageLogger = ResolverUtil.create_instance(usage_logger_class_name,
                                                                 "AGENT_USAGE_LOGGER env var",
                                                                 UsageLogger)
        wrapped: WrappedUsageLogger = WrappedUsageLogger(usage_logger)
        return wrapped




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

import json


class ArgumentAssigner:
    """
    Class which puts the text together for passing function arguments
    information from one agent to the next.
    """

    def __init__(self, properties: Dict[str, Any]):
        """
        Constructor

        :param properties: The dictionary of function properties to fulfill,
                as described in the callee agent spec.
        """
        self.properties: Dict[str, Any] = properties

    def assign(self, arguments: Dict[str, Any]) -> List[str]:
        """
        :param arguments: The arguments dictionary with the values as determined
                by the calling agent.
        :return: A List of text that describes the values of each argument,
                suitable for transmitting to the chat stream of another agent.
        """
        assignments: List[str] = []

        # Start to build the list of assignments, with one sentence for each argument
        for args_name, args_value in arguments.items():

            # Skip if the value of the argument is None or empty
            if args_value is None:
                continue

            # Get argument value type from properties if possible
            args_value_type: str = None
            if self.properties:
                atttribute: Dict[str, Any] = self.properties.get(args_name)
                # Skip if the argument name from llm does not match with that of properties
                if not atttribute:
                    continue
                args_value_type = atttribute.get("type")
            args_value_str: str = self.get_args_value_as_string(args_value, args_value_type)

            # No specific attribution text, so we make up a boilerplate
            # one where it give the arg name <is/are> and the value.

            # Figure out the attribution verb for singular vs plural
            assignment_verb: str = "is"
            if args_value_type == "array" or isinstance(args_value, list):
                assignment_verb = "are"

            # Put together the assignment statement
            assignment: str = f"The {args_name} {assignment_verb} {args_value_str}."

            assignments.append(assignment)

        return assignments

    def get_args_value_as_string(self, args_value: Any, value_type: str = None) -> str:
        """
        Get the string value of the value provided in the arguments
        """
        args_value_str: str = None

        if value_type == "dict" or isinstance(args_value, dict):
            args_value_str = json.dumps(args_value)
            # Strip the begin/end braces as gpt-4o doesn't like them.
            # This means that anything within the json-y braces for a dictionary
            # value gets interpreted as "this is an input value that has
            # to come from the code" when that is not the case at all.
            # Unclear why this is an issue with gpt-4o and not gpt-4-turbo.
            args_value_str = args_value_str[1:-1]

        elif value_type == "array" or isinstance(args_value, list):
            str_values = []
            for item in args_value:
                item_str: str = self.get_args_value_as_string(item)
                str_values.append(item_str)
            args_value_str = ", ".join(str_values)

        elif value_type == "string":
            # For a long time, this had been:
            #       args_value_str = f'"{args_value}"'
            # ... but as of 6/19/25 we are experimenting with new quoting
            #   in an attempt to reduce crazy JSON escaping
            args_value_str = f"'{args_value}'"
            # Per https://github.com/langchain-ai/langchain/issues/1660
            # We need to use double curly braces in order to pass values
            # that actually have curly braces in them so they will not
            # be mistaken for string placeholders for input.
            args_value_str = args_value_str.replace("{", "{{")
            args_value_str = args_value_str.replace("}", "}}")

        else:
            args_value_str = str(args_value)

        return args_value_str




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT


class Startable:
    """
    Interface for objects which have a specific starting phase.
    """

    def start(self):
        """
        Perform start up.
        """
        raise NotImplementedError

    def stop(self):
        """
        Perform steps to stop/shut-down
        By default this does nothing
        """




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from asyncio import run
from os import environ

from neuro_san.interfaces.usage_logger import UsageLogger


class WrappedUsageLogger(UsageLogger):
    """
    Implementation of the UsageLogger interface that wraps another UsageLogger
    and whose log_usage method makes sure sloppier inputs adhere better to the
    UsageLogger contract.
    """

    def __init__(self, wrapped: UsageLogger):
        """
        Constructor

        :param wrapped: The UsageLogger instance that is wrapped.
                        Can be None and log_usage() will handle that.
        """
        self.wrapped: UsageLogger = wrapped

    async def log_usage(self, token_dict: Dict[str, Any], request_metadata: Dict[str, Any]):
        """
        Logs the token usage for external capture.

        :param token_dict: A dictionary that describes overall token usage for a completed request.

                For each class of LLM (more or less equivalent to an LLM provider), there will
                be one key whose value is a dictionary with some other keys:

                Relevant keys include:
                    "completion_tokens" - Integer number of tokens generated in response to LLM input
                    "prompt_tokens" - Integer number of tokens that provide input to an LLM
                    "time_taken_in_seconds" - Float describing the total wall-clock time taken for the request.
                    "total_cost" -  An estimation of the cost in USD of the request.
                                    This number is to be taken with a grain of salt, as these estimations
                                    can come from model costs from libraries instead of directly from
                                    providers.
                    "total_tokens" - Total tokens used for the request.

                More keys can appear, but should not be counted on.
                The ones listed above contain potentially salient information for usage logging purposes.

        :param request_metadata: A dictionary of filtered request metadata whose keys contain
                identifying information for the usage log.
        """
        if self.wrapped is None:
            # Nothing to report
            return

        if token_dict is None:
            # Nothing to report
            return

        compliant_token_dict: Dict[str, Any] = self.make_compliant_token_dict(token_dict)
        minimal_metadata: Dict[str, Any] = self.minimize_metadata(request_metadata)

        await self.wrapped.log_usage(compliant_token_dict, minimal_metadata)

    def synchronous_log_usage(self, token_dict: Dict[str, Any], request_metadata: Dict[str, Any]):
        """
        Logs the token usage for external capture.
        See comments for log_usage() above.
        """
        run(self.log_usage(token_dict, request_metadata))

    def make_compliant_token_dict(self, token_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param token_dict: The token dictionary to make compliant if it is not already
        :return: A token dictionary compliant to the UsageLogger interface
        """
        compliant: Dict[str, Any] = token_dict

        if "total_tokens" in token_dict.keys():
            # We have a raw token dictionary without any attribution to the LLM.
            # Make the dictionary compliant
            compliant = {
                "all": token_dict
            }

        return compliant

    def minimize_metadata(self, request_metadata: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_metadata: The raw request metadata dictionary that could easily contain
                    more keys than we want to send to the UsageLogger.
        :return: A minimized dictionary that only sends the keys we need to the UsageLogger.
                The idea is that this prevents the UsageLogger from getting potentially
                sensitive information it shouldn't really have.

                If the requested keys in the metadata are not there, they will also not appear
                in the returned minimized dictionary.
        """
        minimized: Dict[str, Any] = {}

        # Try getting the value from the more specific env var before falling back to the
        # other env var.
        keys_string: str = environ.get("AGENT_USAGE_LOGGER_METADATA",
                                       environ.get("AGENT_FORWARDED_REQUEST_METADATA"))

        if keys_string is None or len(keys_string) == 0:
            return minimized

        keys: List[str] = keys_string.split(" ")
        for key in keys:
            if key is None or len(key) == 0:
                # Skip any empty key split from the list. Allows for multi-spaces.
                continue
            value: Any = request_metadata.get(key)
            if value is not None:
                minimized[key] = value

        return minimized




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.internals.interfaces.error_formatter import ErrorFormatter


class StringErrorFormatter(ErrorFormatter):
    """
    Implementation of ErrorFormatter interface which compiles error information
    into a string.   See parent interface for more details.
    """

    def format(self, agent_name: str, message: str, details: str = None) -> str:
        """
        Format an error message

        :param agent_name: A string describing the name of the agent experiencing
                the error.
        :param message: The specific message describing the error occurrence.
        :param details: An optional string describing further details of how/where the
                error occurred.  Think: traceback.
        :return: String encapsulation and/or filter of the error information
                presented in the arguments.
        """
        error: str = f"Error from {agent_name}: {message}"
        if details is not None:
            error = f"{error} ({details})"
        return error




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Dict

from janus import Queue

from leaf_common.asyncio.asyncio_executor_pool import AsyncioExecutorPool

from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.internals.network_providers.expiring_agent_network_storage import ExpiringAgentNetworkStorage
from neuro_san.service.utils.server_status import ServerStatus


class ServerContext:
    """
    Class that contains global-ish state for each instance of a server.
    """

    def __init__(self):
        """
        Constructor.
        """
        self.server_status: ServerStatus = None
        self.executor_pool = AsyncioExecutorPool(reuse_mode=True)
        self.queues: Queue[AsyncCollatingQueue] = Queue()

        # Dictionary is string key (describing scope) to AgentNetworkStorage grouping.
        self.network_storage_dict: Dict[str, AgentNetworkStorage] = {
            "protected": AgentNetworkStorage(),
            "public": AgentNetworkStorage(),
            "temp": ExpiringAgentNetworkStorage()
        }

    def get_executor_pool(self) -> AsyncioExecutorPool:
        """
        :return: The AsyncioExecutorPool
        """
        return self.executor_pool

    def set_server_status(self, server_status: ServerStatus):
        """
        Sets the server status
        """
        self.server_status = server_status

    def get_server_status(self) -> ServerStatus:
        """
        :return: The ServerStatus
        """
        return self.server_status

    def get_network_storage_dict(self) -> Dict[str, AgentNetworkStorage]:
        """
        :return: The Network Storage dictionary
        """
        return self.network_storage_dict

    def get_queues(self) -> Queue[AsyncCollatingQueue]:
        """
        :return: The janus Queue of queues for temporary agent deployment
        """
        return self.queues

    def no_queues(self):
        """
        Resets the queues to None as a signal to other parts of code base
        that we don't need Reservationists
        """
        self.queues = None




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Iterator

import contextlib
import copy
import json
import uuid

from janus import Queue

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor
from leaf_common.asyncio.asyncio_executor_pool import AsyncioExecutorPool
from leaf_common.utils.atomic_counter import AtomicCounter

from leaf_server_common.server.request_logger import RequestLogger

from neuro_san.interfaces.reservationist import Reservationist
from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.run_context.factory.master_toolbox_factory import MasterToolboxFactory
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.service.generic.agent_server_logging import AgentServerLogging
from neuro_san.service.generic.chat_message_converter import ChatMessageConverter
from neuro_san.service.generic.service_agent_reservationist import ServiceAgentReservationist
from neuro_san.service.usage.usage_logger_factory import UsageLoggerFactory
from neuro_san.service.usage.wrapped_usage_logger import WrappedUsageLogger
from neuro_san.service.utils.server_context import ServerContext
from neuro_san.session.direct_agent_session import DirectAgentSession
from neuro_san.session.external_agent_session_factory import ExternalAgentSessionFactory
from neuro_san.session.session_invocation_context import SessionInvocationContext

# A list of methods to not log requests for
# Some of these can be way too chatty
DO_NOT_LOG_REQUESTS = [
]


# pylint: disable=too-many-instance-attributes
class AgentService:
    """
    A base implementation of the Neuro-San Agent Service,
    independent of target transport protocol.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self,
                 request_logger: RequestLogger,
                 security_cfg: Dict[str, Any],
                 agent_name: str,
                 agent_network_provider: AgentNetworkProvider,
                 server_logging: AgentServerLogging,
                 server_context: ServerContext):
        """
        Set the gRPC interface up for health checking so that the service
        will be opened to callers when the mesh sees it operational, if this
        is not done the mesh will treat the service instance as non functional

        :param request_logger: The instance of the RequestLogger that helps
                    keep track of stats
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  If None, uses insecure channel.
        :param agent_name: The agent name for the service
        :param agent_network_provider: The AgentNetworkProvider to use for the session.
        :param server_logging: An AgentServerLogging instance initialized so that
                        spawned asyncrhonous threads can also properly initialize
                        their logging.
        :param server_context: The ServerContext containing global-ish state
        """
        self.request_logger = request_logger
        self.security_cfg = security_cfg
        self.server_logging: AgentServerLogging = server_logging

        self.agent_network_provider: AgentNetworkProvider = agent_network_provider
        self.agent_name: str = agent_name
        self.request_counter = AtomicCounter()

        # Stuff needed for ServiceAgentReservationist
        self.queues: Queue[AsyncCollatingQueue] = server_context.get_queues()

        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        config: Dict[str, Any] = agent_network.get_config()
        self.llm_factory: ContextTypeLlmFactory = MasterLlmFactory.create_llm_factory(config)
        self.toolbox_factory: ContextTypeToolboxFactory = MasterToolboxFactory.create_toolbox_factory(config)
        self.async_executor_pool: AsyncioExecutorPool = server_context.get_executor_pool()
        # Load once
        self.llm_factory.load()
        self.toolbox_factory.load()

    def get_request_count(self) -> int:
        """
        :return: The number of currently active requests
        """
        return self.request_counter.get_count()

    def function(self, request_dict: Dict[str, Any],
                 request_metadata: Dict[str, Any],
                 context: Any) \
            -> Dict[str, Any]:
        """
        Allows a client to get the outward-facing function for the agent
        served by this service.

        :param request_dict: a FunctionRequest dictionary
        :param request_metadata: request metadata
        :param context: a service request context object
        :return: a FunctionResponse dictionary
        """
        self.request_counter.increment()
        request_log = None
        log_marker: str = "function request"
        service_logging_dict: Dict[str, str] = {
            "request_id": f"server-{uuid.uuid4()}"
        }
        if "Function" not in DO_NOT_LOG_REQUESTS:
            request_log = self.request_logger.start_request(f"{self.agent_name}.Function",
                                                            log_marker, context,
                                                            service_logging_dict)

        # Get the metadata to forward on to another service
        metadata: Dict[str, str] = copy.copy(service_logging_dict)
        metadata.update(request_metadata)

        # Delegate to Direct*Session
        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        session = DirectAgentSession(agent_network=agent_network,
                                     invocation_context=None,
                                     metadata=metadata,
                                     security_cfg=self.security_cfg)
        response_dict = session.function(request_dict)

        if request_log is not None:
            self.request_logger.finish_request(f"{self.agent_name}.Function", log_marker, request_log)

        self.request_counter.decrement()
        return response_dict

    def connectivity(self, request_dict: Dict[str, Any],
                     request_metadata: Dict[str, Any],
                     context: Any) \
            -> Dict[str, Any]:
        """
        Allows a client to get connectivity information for the agent
        served by this service.

        :param request_dict: a ChatRequest dictionary
        :param request_metadata: request metadata
        :param context: a service request context object
        :return: a ConnectivityResponse dictionary
        """
        self.request_counter.increment()
        request_log = None
        log_marker: str = "connectivity request"
        service_logging_dict: Dict[str, str] = {
            "request_id": f"server-{uuid.uuid4()}"
        }
        if "Connectivity" not in DO_NOT_LOG_REQUESTS:
            request_log = self.request_logger.start_request(f"{self.agent_name}.Connectivity",
                                                            log_marker, context,
                                                            service_logging_dict)

        # Get the metadata to forward on to another service
        metadata: Dict[str, str] = copy.copy(service_logging_dict)
        metadata.update(request_metadata)

        # Delegate to Direct*Session
        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        session = DirectAgentSession(agent_network=agent_network,
                                     invocation_context=None,
                                     metadata=metadata,
                                     security_cfg=self.security_cfg)
        response_dict = session.connectivity(request_dict)

        if request_log is not None:
            self.request_logger.finish_request(f"{self.agent_name}.Connectivity", log_marker, request_log)

        self.request_counter.decrement()
        return response_dict

    # pylint: disable=too-many-locals
    def streaming_chat(self, request_dict: Dict[str, Any],
                       request_metadata: Dict[str, Any],
                       context: Any) \
            -> Iterator[Dict[str, Any]]:
        """
        Initiates or continues the agent chat with the session_id
        context in the request.

        :param request_dict: a ChatRequest dictionary
        :param request_metadata: request metadata
        :param context: a service request context object
        :return: an iterator for (eventually) returned responses dictionaries
        """
        self.request_counter.increment()
        request_log = None
        user_text: str = request_dict.get("user_message", {}).get("text", "")
        log_marker = f"'{user_text}'"
        service_logging_dict: Dict[str, str] = {
            "request_id": f"server-{uuid.uuid4()}"
        }
        if "StreamingChat" not in DO_NOT_LOG_REQUESTS:
            request_log = self.request_logger.start_request(f"{self.agent_name}.StreamingChat",
                                                            log_marker, context,
                                                            service_logging_dict)

        # Get the metadata to forward on to another service
        metadata: Dict[str, str] = copy.copy(service_logging_dict)
        metadata.update(request_metadata)
        if metadata.get("request_id") is None:
            metadata["request_id"] = service_logging_dict.get("request_id")

        # Create a reservationist
        reservationist: Reservationist = None
        if self.queues is not None:
            reservationist = ServiceAgentReservationist()
            self.queues.sync_q.put(reservationist.get_queue())

        # Prepare
        factory = ExternalAgentSessionFactory(use_direct=False)
        invocation_context = SessionInvocationContext(
            factory,
            self.async_executor_pool,
            self.llm_factory,
            self.toolbox_factory,
            metadata,
            reservationist)
        invocation_context.start()

        # Set up logging inside async thread
        # Prefer any request_id from the client over what we generated on the server.
        executor: AsyncioExecutor = invocation_context.get_asyncio_executor()
        _ = executor.submit(None, self.server_logging.setup_logging, metadata, metadata.get("request_id"))

        # Delegate to Direct*Session
        agent_network: AgentNetwork = self.agent_network_provider.get_agent_network()
        session = DirectAgentSession(agent_network=agent_network,
                                     invocation_context=invocation_context,
                                     metadata=metadata,
                                     security_cfg=self.security_cfg)
        # Get our args in order to pass to grpc-free session level
        response_dict_iterator: Iterator[Dict[str, Any]] = session.streaming_chat(request_dict)

        # See if we want to put the request dict in the response
        chat_filter_dict: Dict[str, Any] = {}
        chat_filter_dict = request_dict.get("chat_filter", chat_filter_dict)
        chat_filter_type: str = chat_filter_dict.get("chat_filter_type", "MINIMAL")

        try:
            for response_dict in response_dict_iterator:
                # Prepare chat message for output:
                response_dict = ChatMessageConverter().to_dict(response_dict)
                # Do not return the request when the filter is MINIMAL
                if chat_filter_type != "MINIMAL":
                    response_dict["request"] = request_dict
                yield response_dict
        finally:
            # Put iterator cleanup logic in "finally" part of try-except block;
            # this way we guarantee that underlying response_dict_iterator will be closed
            # whether we finish consuming its data stream normally
            # OR we are interrupted downstream
            # and have special "GeneratorExit" exception delivered to us.
            request_reporting: Dict[str, Any] = invocation_context.get_request_reporting()
            # Properly close our iterator:
            if response_dict_iterator is not None:
                with contextlib.suppress(Exception):
                    response_dict_iterator.close()
            # Ensure that our SessionInvocationContext is always closed,
            # even if iterator is interrupted.
            invocation_context.close()
            invocation_context = None

        # Maybe report token accounting to a UsageLogger
        token_dict: Dict[str, Any] = request_reporting.get("token_accounting")
        if token_dict is not None:
            usage_logger: WrappedUsageLogger = UsageLoggerFactory.create_usage_logger()
            usage_logger.synchronous_log_usage(token_dict, request_metadata)

        # Iterator has finally signaled that there are no more responses to be had.
        # Log that we are done.
        if request_log is not None:
            reporting: str = None
            if request_reporting is not None:
                reporting = json.dumps(request_reporting, indent=4, sort_keys=False)
            request_log.metrics("Request reporting: %s", reporting)
            self.request_logger.finish_request(f"{self.agent_name}.StreamingChat", log_marker, request_log)

        self.request_counter.decrement()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT


class Reservation:
    """
    A data structure containing information about an id procured for a specific amount of time.
    """

    def __init__(self, lifetime_in_seconds: float):
        """
        Constructor

        :param lifetime_in_seconds: The number of seconds the reservation is allowed to exist.
        """
        # This id is to be assigned by the implementations however they see fit.
        # We recommend uuid4().
        self.id: str = None
        self.lifetime_in_seconds: float = lifetime_in_seconds
        self.expiration_time_in_seconds: float = 0.0

    def get_reservation_id(self) -> str:
        """
        :return: The id associated with the reservation instance
        """
        return self.id

    def get_lifetime_in_seconds(self) -> float:
        """
        :return: The lifetime in seconds associated with the reservation
        """
        return self.lifetime_in_seconds

    def get_expiration_time_in_seconds(self) -> float:
        """
        :return: The expiration time in seconds since the epoch, ala time.time().
        """
        return self.expiration_time_in_seconds

    def get_url(self) -> str:
        """
        :return: A url associated with the reservation.
                 Can be None if this is not an option for what we are reserving.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.parsers.structure.structure_parser import StructureParser
from neuro_san.internals.parsers.structure.structure_parser_factory import StructureParserFactory


class FirstAvailableStructureParser(StructureParser):
    """
    StructureParser implementation that takes a list of possible formats
    and does structure parsing on the first one that matches for the text.
    """

    def __init__(self, structure_formats: List[str]):
        """
        Constructor

        :param structure_formats: A List of strings containing potential parse formats to match
        """
        super().__init__()
        self.structure_formats: List[str] = structure_formats
        self.factory = StructureParserFactory()

    def parse_structure(self, content: str) -> Dict[str, Any]:
        """
        Parse the single string content for any signs of structure

        :param content: The string to parse for structure
        :return: A dictionary structure that was embedded in the content.
                Will return None if no parseable structure is detected.
        """
        structure: Dict[str, Any] = None

        for structure_format in self.structure_formats:

            structure_parser: StructureParser = self.factory.create_structure_parser(structure_format)
            if structure_parser is None:
                # Format did not match anything we know how to parse. Keep looking.
                continue

            structure = structure_parser.parse_structure(content)
            if structure is not None:
                # We only do the first that matches
                self.remainder = structure_parser.get_remainder()
                break

        return structure




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See comments in class description
"""

import logging
from typing import Any
from typing import Dict
from threading import Lock
import grpc


class DynamicAgentRouter(grpc.GenericRpcHandler):
    """
    Class maintains a dynamic map of gRPC services exposed by a server.
    Each gRPC service represents an agent available to clients.
    Instance of this class is also a generic gRPC handler,
    routing incoming gRPC request to appropriate API method handler
    depending on service name.
    """

    instance = None

    def __init__(self):
        """
        Constructor.
        """
        self.agents_table: Dict[str, Any] = {}
        self.table_lock: Lock = Lock()
        self.logger = logging.getLogger(self.__class__.__name__)

    def add_service(self, service_name: str, handlers: Dict[str, Any]):
        """
        Add service with its API handlers to this router.
        :param service_name: gRPC-formatted service name;
        :param handlers: table of API handlers to register for this service.
        """
        with self.table_lock:
            self.agents_table[service_name] = handlers
        self.logger.info("Added service handlers for %s", service_name)

    def remove_service(self, service_name: str):
        """
        Remove service from this router.
        :param service_name: gRPC-formatted service name;
        """
        with self.table_lock:
            self.agents_table.pop(service_name, None)
        self.logger.info("Removed service handlers for %s", service_name)

    def service(self, handler_call_details: grpc.HandlerCallDetails):
        """
        Service incoming gRPC request.
        :param handler_call_details: gRPC incoming call details.
        :return: gRPC method handler for this call. This handler will be invoked
            elsewhere by gRPC machinery.
        """
        full_method: str = "unknown"
        try:
            full_method = handler_call_details.method  # e.g. "/my.Service/SomeMethod"
            _, service_name, method_name = full_method.split("/", 2)
            with self.table_lock:
                method_handlers = self.agents_table.get(service_name, None)
            if method_handlers:
                return method_handlers.get(method_name, None)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.logger.error("Failed to execute %s - %s", full_method, str(exc))
        # Request not handled, in this case gRPC stack will raise UNIMPLEMENTED exception
        # and http server will return error code 500 with details: "Method not found!"
        return None




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Union

from copy import copy
from datetime import datetime
from os import environ

from leaf_common.config.file_of_class import FileOfClass
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor
from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence
from leaf_common.time.timeout import Timeout

from neuro_san.client.agent_session_factory import AgentSessionFactory
from neuro_san.client.streaming_input_processor import StreamingInputProcessor
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.message_processing.basic_message_processor import BasicMessageProcessor
from neuro_san.session.direct_agent_session import DirectAgentSession
from neuro_san.test.driver.assert_capture import AssertCapture
from neuro_san.test.evaluators.agent_evaluator_factory import AgentEvaluatorFactory
from neuro_san.test.interfaces.agent_evaluator import AgentEvaluator
from neuro_san.test.interfaces.assert_forwarder import AssertForwarder


class DataDrivenAgentTestDriver:
    """
    Class which manages the execution of a single data-driven test case
    specified as a hocon file.
    """

    TEST_KEYS: List[str] = ["text", "structure", "sly_data"]

    def __init__(self, asserts: AssertForwarder, fixtures: FileOfClass = None):
        """
        Constructor
        :param asserts: The AssertForwarder instance to use to integrate failures
                        back into the test system.
        :param fixtures: Optional path to the fixtures root.
        """
        self.asserts_basis: AssertForwarder = asserts
        self.fixtures: FileOfClass = fixtures

    # pylint: disable=too-many-locals
    def one_test(self, hocon_file: str):
        """
        Use a single hocon file in the fixtures as a test case"

        :param hocon_file: The name of the hocon from the fixtures directory.
        """
        test_case: Dict[str, Any] = self.parse_hocon_test_case(hocon_file)

        agent: str = test_case.get("agent")
        self.asserts_basis.assertIsNotNone(agent)

        # Set up any global test timeout.
        timeouts: List[Timeout] = []
        timeout_in_seconds: float = test_case.get("timeout_in_seconds", None)
        if timeout_in_seconds is not None:
            test_timeout = Timeout(name=agent)
            test_timeout.set_limit_in_seconds(timeout_in_seconds)
            timeouts.append(test_timeout)

        # Get the success ratio
        success_ratio: str = test_case.get("success_ratio", "1/1")
        self.asserts_basis.assertIn("/", success_ratio)

        # Find the integer components of the success ratio
        success_split: List[str] = success_ratio.split("/")
        num_need_success: int = int(success_split[0])
        num_iterations: int = int(success_split[-1])

        # Put some bounds on the number of iterations
        num_iterations = max(1, num_iterations)
        num_need_success = min(num_need_success, num_iterations)

        # Capture asserts for each iteration
        iteration_asserts: List[AssertCapture] = []

        # Loop through each iteration, capturing any asserts.
        num_successful: int = 0
        for index in range(num_iterations):

            _ = index

            # Capture the asserts for this iteration and add it to the list for later
            assert_capture = AssertCapture(self.asserts_basis)
            iteration_asserts.append(assert_capture)

            # Perform a single iteration of the test.
            self.one_iteration(test_case, assert_capture, timeouts)

            # Update our counter if this iteration is successful
            asserts: List[AssertionError] = assert_capture.get_asserts()
            if len(asserts) > 0:
                # Not successful
                continue

            num_successful += 1
            if num_successful == num_need_success:
                # Don't do more tests than we actually need to
                break

        # Don't bother reporting any asserts if we have met our success ratio.
        # Return early to pass this test.
        if num_successful >= num_need_success:
            return

        # Find the first assert that fails and use it to fail this test
        for assert_capture in iteration_asserts:
            asserts: List[AssertionError] = assert_capture.get_asserts()
            if len(asserts) > 0:
                one_assert: AssertionError = asserts[0]
                message: str = f"""
{num_successful} of {num_iterations} iterations on agent {agent} were successful.
Need at least {num_need_success} to consider {hocon_file} test to be successful.
"""
                raise AssertionError(message) from one_assert

    # pylint: disable=too-many-locals
    def one_iteration(self, test_case: Dict[str, Any], asserts: AssertForwarder, timeouts: List[Timeout]):
        """
        Perform a single iteration on the test case.

        :param test_case: The dictionary describing the data-driven test case
        :param asserts: The AssertForwarder to send asserts to.
        :param timeouts: A list of timeout objects to check
        """

        # Get the agent to use
        agent: str = test_case.get("agent")

        # Get the connection type
        connections: Union[List[str], str] = test_case.get("connections")
        if connections is None:
            # Assume direct if not specified
            connections = ["direct"]
        elif isinstance(connections, str):
            # Make single strings into a list for consistent parsing
            connections = [connections]
        asserts.assertIsInstance(connections, List)
        asserts.assertGreater(len(connections), 0)

        # Collect the interations to test for
        empty: List[Any] = []
        interactions: List[Dict[str, Any]] = test_case.get("interactions", empty)
        asserts.assertGreater(len(interactions), 0)

        # Collect other session information
        use_direct: bool = test_case.get("use_direct", False)
        timeout_in_seconds: float = test_case.get("timeout_in_seconds", None)
        metadata: Dict[str, Any] = test_case.get("metadata", None)
        if metadata is None:
            # Use a default from the user's environment to at least let
            # a server know who is doing the querying.
            metadata = {
                "user_id": environ.get("USER")
            }

        for connection in connections:

            session: AgentSession = AgentSessionFactory().create_session(
                    connection,
                    agent,
                    use_direct=use_direct,
                    metadata=metadata,
                    connect_timeout_in_seconds=timeout_in_seconds)
            chat_context: Dict[str, Any] = None
            for interaction in interactions:

                if isinstance(session, DirectAgentSession):
                    session.reset()

                chat_context = self.interact(agent, session, interaction, chat_context, asserts,
                                             timeouts)

    def parse_hocon_test_case(self, hocon_file: str) -> Dict[str, Any]:
        """
        Use a single hocon file in the fixtures as a test case"

        :param hocon_file: The name of the hocon from the fixtures directory.
        """
        test_path: str = hocon_file
        if self.fixtures is not None:
            test_path = self.fixtures.get_file_in_basis(hocon_file)
        hocon = EasyHoconPersistence(must_exist=True)
        test_case: Dict[str, Any] = hocon.restore(file_reference=test_path)
        return test_case

    # pylint: disable=too-many-locals,too-many-arguments,too-many-positional-arguments
    def interact(self, agent: str, session: AgentSession, interaction: Dict[str, Any],
                 chat_context: Dict[str, Any], asserts: AssertForwarder,
                 timeouts: List[Timeout]) -> Dict[str, Any]:
        """
        Interact with an agent and evaluate its output

        :param session: The AgentSession to work with
        :param interaction: The interaction dictionary to base evalaution off of.
        :param chat_context: The chat context to use with the interaction (if any)
        :param asserts: The AssertForwarder to send asserts to.
        :param timeouts: A list of timeout objects to check
        """
        _ = agent       # For now
        empty: Dict[str, Any] = {}

        # Shallow copy what we already have in timeouts
        use_timeouts: List[Timeout] = copy(timeouts)

        # Prepare the processor
        now = datetime.now()
        datestr: str = now.strftime("%Y-%m-%d-%H:%M:%S")
        thinking_dir: str = f"/tmp/agent_test/{datestr}_agent"
        input_processor = StreamingInputProcessor("", None, session, thinking_dir)
        processor: BasicMessageProcessor = input_processor.get_message_processor()

        # Prepare the request
        text: str = interaction.get("text")
        sly_data: str = interaction.get("sly_data")
        chat_filter: Dict[str, Any] = {
            "chat_filter_type": interaction.get("chat_filter", "MINIMAL")
        }
        request: Dict[str, Any] = input_processor.formulate_chat_request(text, sly_data, chat_context, chat_filter)

        # Prepare any interaction timeout
        if interaction.get("timeout_in_seconds") is not None:
            interaction_timeout = Timeout(name=text)
            interaction_timeout.set_limit_in_seconds(interaction.get("timeout_in_seconds"))
            use_timeouts.append(interaction_timeout)

        # Call streaming_chat()
        chat_responses: Generator[Dict[str, Any], None, None] = session.streaming_chat(request)
        for chat_response in chat_responses:
            message = chat_response.get("response", empty)
            processor.process_message(message, chat_response.get("type"))
            self.check_timeouts(use_timeouts)

        self.check_timeouts(use_timeouts)

        # Evaluate response
        response: Dict[str, Any] = interaction.get("response", empty)
        response_extractor = DictionaryExtractor(response)
        self.test_response_keys(processor, response_extractor, self.TEST_KEYS, asserts, use_timeouts)
        self.check_timeouts(use_timeouts)

        # See how we should continue the conversation
        return_chat_context: Dict[str, Any] = None
        if interaction.get("continue_conversation", True):
            return_chat_context = processor.get_chat_context()

        return return_chat_context

    def test_response_keys(self, processor: BasicMessageProcessor,
                           response_extractor: DictionaryExtractor,
                           keys: List[str],
                           asserts: AssertForwarder,
                           timeouts: List[Timeout]):
        """
        Tests the given response keys

        :param processor: The BasicMessageProcessor instance to query results from.
        :param response_extractor: The DictionaryExtractor for the test structure from the test hocon file.
        :param keys: The response keys to test
        :param asserts: The AssertForwarder to send asserts to.
        :param timeouts: A list of timeout objects to check
        """
        deeper_test_keys: List[str] = []

        for test_key in keys:

            test_key_value: Dict[str, Any] = response_extractor.get(test_key)
            if test_key_value is None:
                # Got nothing for test_key. Nothing to see here. Please move along.
                continue

            if isinstance(test_key_value, Dict):
                # The value refers to a deeper dictionary test
                for deeper_key in test_key_value.keys():
                    deeper_test_keys.append(f"{test_key}.{deeper_key}")
            else:
                # The last part of the test_key refers to a specific evaluator type.
                split: List[str] = test_key.split(".")
                evaluator_type: str = split[-1]            # Last component of .-delimited key
                verify_key: str = ".".join(split[:-1])      # All but last component of .-delimited key
                evaluator: AgentEvaluator = AgentEvaluatorFactory.create_evaluator(asserts,
                                                                                   evaluator_type)
                if evaluator is not None:
                    evaluator.evaluate(processor, verify_key, test_key_value)
                    self.check_timeouts(timeouts)

        # Recurse if there are further dictionary specs to dive into
        if len(deeper_test_keys) > 0:
            self.test_response_keys(processor, response_extractor, deeper_test_keys, asserts, timeouts)

    def check_timeouts(self, timeouts: List[Timeout]):
        """
        :param timeouts: A list of timeout objects to check
        """
        for one_timeout in timeouts:
            Timeout.check_if_not_none(one_timeout)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT


from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import os

from langchain_core.tools.base import BaseTool
from langchain_community.agent_toolkits.base import BaseToolkit
from pydantic import BaseModel

from leaf_common.config.dictionary_overlay import DictionaryOverlay
from leaf_common.config.resolver import Resolver

from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.run_context.langchain.toolbox.toolbox_info_restorer import ToolboxInfoRestorer
from neuro_san.internals.run_context.langchain.util.argument_validator import ArgumentValidator


class ToolboxFactory(ContextTypeToolboxFactory):
    """
    A factory class for creating instances of various tools defined in the toolbox.

    This class provides an interface to instantiate different tools based on the specified langchain base tools
    and predefined coded tools.

    This approach standardizes tool creation and simplifies integration with agents requiring predefined tools.

    ### Extending the Class

        To integrate additional tools, add a tool configuration file in JSON or HOCON format
        and set its path to the environment variable "AGENT_TOOLBOX_INFO_FILE".

        The configuration should follow this structure

        for langchain's tools:
        - The tool name serves as a key.
        - The corresponding value should be a dictionary with:
        - "class": The fully qualified class name of the tool.
        - "args": A dictionary of arguments required for the tool's initialization,
            which may include nested class configurations.

        for coded tools:
        - The tool name serves as a key.
        - The corresponding value should be a dictionary with:
        - "class": Module and class in the format of tool_module.ClassName where tool_module is in
                    AGENT_TOOL_PATH or neuro_san/coded_tools.
        - "description": When and how to use the tool.
        - "parameters": Information on arguments of the tool.
            See "parameters" in https://github.com/cognizant-ai-lab/neuro-san/blob/main/docs/agent_hocon_reference.md

        The default toolbox config file can be seen at
        "neuro_san/internals/run_context/langchain/toolbox/toolbox_info.hocon"
    """

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """
        Constructor

        :param config: The config dictionary which may or may not contain
                       keys for the context_type and toolbox_info_file
        """
        self.toolbox_infos: Dict[str, Any] = {}
        self.overlayer = DictionaryOverlay()

        # Get user toolbox info file path with the following priority:
        # 1. "toolbox_info_file" from agent network hocon
        # 2. "AGENT_TOOLBOX_INFO_FILE" from environment variable
        if config:
            raw_toolbox_info_file: str = (
                config.get("toolbox_info_file")
                or os.getenv("AGENT_TOOLBOX_INFO_FILE")
            )
        else:
            raw_toolbox_info_file = os.getenv("AGENT_TOOLBOX_INFO_FILE")

        if raw_toolbox_info_file is not None and not isinstance(raw_toolbox_info_file, str):
            raise TypeError(
                "The values of 'toolbox_info_file' and "
                "the 'AGENT_TOOLBOX_INFO_FILE' environment variable must be strings. "
                f"Got {type(raw_toolbox_info_file).__name__} instead."
            )

        self.toolbox_info_file: str = raw_toolbox_info_file

    def load(self):
        """
        Loads the base tool information from hocon files.
        """
        restorer = ToolboxInfoRestorer()
        self.toolbox_infos = restorer.restore()

        # Mix in user-specified toolbox info, if available.
        if self.toolbox_info_file:
            extra_toolbox_infos: Dict[str, Any] = restorer.restore(file_reference=self.toolbox_info_file)
            self.toolbox_infos = self.overlayer.overlay(self.toolbox_infos, extra_toolbox_infos)

    def create_tool_from_toolbox(
            self,
            tool_name: str,
            user_args: Dict[str, Any] = None,
            agent_name: str = None
    ) -> Union[BaseTool, Dict[str, Any], List[BaseTool]]:
        """
        Resolves dependencies and instantiates the requested tool.

        :param tool_name: The name of the tool to instantiate.
        :param user_args: Arguments provided by the user, which override the config file.
        :param agent_name: The name of the agent to prefix each BaseTool's name in BaseToolkit with,
            ensuring tool names are unique and traceable to their agent.
        :return: - Instantiated tool if "class" of tool_name points to a BaseTool class
                 - A list of tools if "class of "tool_name points to a BaseToolkit class.
                 - A dict of tool's "description" and "parameters" if tool_name points to a CodedTool
        """

        # agent_name is required when the tool is used as an internal agent.
        # However, tools from the toolbox could potentially be used as external tools,
        # in which case agent_name may not be needed.

        empty: Dict[str, Any] = {}

        tool_info: Dict[str, Any] = self.toolbox_infos.get(tool_name)
        if not tool_info:
            raise ValueError(f"Tool '{tool_name}' is not defined in {self.toolbox_info_file}.")

        if not isinstance(tool_info, Dict):
            raise ValueError(f"The value for the {tool_name} key must be a dictionary.")

        if "class" not in tool_info:
            raise ValueError(
                "Missing required key: 'class'.\n"
                "Each tool must include a 'class' key:\n"
                "- For Langchain base tools: use the full class path "
                "(e.g., 'langchain_community.tools.bing_search.BingSearchResults')\n"
                "- For shared CodedTools: use 'module.Class' format (e.g., 'websearch.WebSearch')"
            )

        # If "description" in the tool info, then it is a shared coded tool.
        # Return dictionary of tool's description and parameters.
        if "description" in tool_info:
            return tool_info

        # Instantiate the main tool or toolkit class
        tool_class: Type[Any] = self._resolve_class(tool_info.get("class"))
        # Recursively resolve arguments (including wrapper dependencies)
        resolved_args: Dict[str, Any] = self._resolve_args(tool_info.get("args", empty))
        # Merge with user arguments where user_args get the priority
        final_args: Dict[str, Any] = self.overlayer.overlay(resolved_args, user_args) if user_args else resolved_args

        # Use the "from_{tool_name}_api_wrapper" method if available, otherwise the constructor
        callable_obj: Union[Type[BaseTool], Type[BaseToolkit], Callable[..., Any]] = \
            self._get_from_api_wrapper_method(tool_class) or tool_class

        # Validate and instantiate
        ArgumentValidator.check_invalid_args(callable_obj, final_args)
        # Instance can be a BaseTool or a BaseToolkit
        instance: Union[BaseTool, BaseToolkit] = callable_obj(**final_args)

        # If the instantiated class has "get_tools()", assume it's a toolkit and return a list of tools
        if hasattr(instance, "get_tools") and callable(instance.get_tools):
            toolkit: List[BaseTool] = instance.get_tools()
            for tool in toolkit:
                if agent_name:
                    # Prefix the name of the agent to each tool
                    tool.name = f"{agent_name}_{tool.name}"
                # Add "langchain_tool" tags so journal callback can idenitify it
                tool.tags = ["langchain_tool"]
            return toolkit

        if agent_name:
            # Replace langchain tool's name with agent name
            instance.name = agent_name
        # Add "langchain_tool" tags so journal callback can idenitify it
        instance.tags = ["langchain_tool"]
        return instance

    def _resolve_args(self, args: Dict[str, Any]) -> Dict[str, Any]:
        """
        Recursive resolves arguments when there is a wrapper class as an argument,
        otherwise return args as a dictionary.

        :param args: The arguments to resolve.
        :return: A dictionary of resolved arguments.
        """
        empty: Dict[str, Any] = {}

        resolved_args: Dict[str, Any] = {}
        for key, value in args.items():
            if isinstance(value, dict) and "class" in value:
                # If the argument is a class definition, resolve and instantiate it
                nested_class: BaseModel = self._resolve_class(value.get("class"))
                nested_args: Dict[str, Any] = self._resolve_args(value.get("args", empty))
                ArgumentValidator.check_invalid_args(nested_class, nested_args)
                resolved_args[key] = nested_class(**nested_args)
            else:
                # Otherwise, keep primitive values as they are
                resolved_args[key] = value
        return resolved_args

    def _resolve_class(self, class_path: str) -> Type[BaseTool]:
        """
        Uses Resolver to dynamically import a class.

        :param class_path: Full class path (e.g., "package.module.ClassName").
        :return: The resolved class type.
        """
        class_split: List[str] = class_path.split(".")
        if len(class_split) <= 2:
            raise ValueError(
                f"Value in 'class' in {self.toolbox_info_file} must be of the form "
                "'<package_name>.<module_name>.<ClassName>'"
            )

        # Extract module and class details
        packages: List[str] = [".".join(class_split[:-2])]
        class_name: str = class_split[-1]
        resolver = Resolver(packages)

        # Resolve class
        try:
            return resolver.resolve_class_in_module(class_name, module_name=class_split[-2])
        except AttributeError as exception:
            raise ValueError(f"Class {class_path} not found in PYTHONPATH") from exception

    def _get_from_api_wrapper_method(
        self,
        tool_class: Union[Type[BaseTool], Type[BaseToolkit]]
    ) -> Optional[Callable[..., Any]]:
        """
        Get a 'from_{tool_name}_api_wrapper' class method from the tool class if available.

        :param tool_class: BaseTool or BaseToolkit class to check for the method.
        :return: The method if found, None otherwise.
        """
        for attr_name in dir(tool_class):
            if attr_name.startswith("from") and attr_name.endswith("api_wrapper"):
                attr: Callable[..., Any] = getattr(tool_class, attr_name)
                if callable(attr):
                    return attr
        return None

    def get_shared_coded_tool_class(self, tool_name: str) -> str:
        """
        Get class of the shared coded tool

        :param tool_name: The name of the tool
        :return: The class of the coded tool
        """
        tool_info: Dict[str, Any] = self.toolbox_infos.get(tool_name)
        return tool_info.get("class")

    def get_tool_info(self, tool_name: str) -> Dict[str, Any]:
        """
        :param tool_name: The name of the tool.
        :return: The toolbox dictionary entry for the tool name
        """
        tool_info: Dict[str, Any] = self.toolbox_infos.get(tool_name)
        return tool_info




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

import os


class EnvironmentConfiguration:
    """
    Easy policy add on for the get_value_or_env() method for various classes that
    are effected by configuration via dictionary/hocon and/or environment variables.
    """

    @staticmethod
    def get_value_or_env(config: Dict[str, Any], key: str, env_key: str,
                         none_obj: Any = None) -> Any:
        """
        :param config: The config dictionary to search
        :param key: The key for the config to look for
        :param env_key: The os.environ key whose value should be gotten if either
                        the key does not exist or the value for the key is None
        :param none_obj:  An optional object instance to test.
                          If present this method will return None, implying
                          that some other external object/mechanism is supplying the values.
        """
        if none_obj is not None:
            return None

        value = None
        if config is not None:
            value = config.get(key)

        if value is None and env_key is not None:
            value = os.getenv(env_key)

        return value




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import Generator

import asyncio
import contextlib
import json
import tornado

from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.http.handlers.base_request_handler import BaseRequestHandler


class StreamingChatHandler(BaseRequestHandler):
    """
    Handler class for neuro-san streaming chat API call.
    """

    async def stream_out(self,
                         generator: Generator[Dict[str, Any], None, None]) -> int:
        """
        Process streaming out generator output to HTTP connection.
        :param generator: async chat response generator
        :return: number of chat responses streamed out.
        """
        # Set up headers for chunked response
        self.set_header("Content-Type", "application/json-lines")
        self.set_header("Transfer-Encoding", "chunked")
        # Flush headers immediately
        flush_ok: bool = await self.do_flush()
        if not flush_ok:
            return 0

        sent_out: int = 0
        async for result_dict in generator:
            result_str: str = json.dumps(result_dict) + "\n"
            self.write(result_str)
            flush_ok = await self.do_flush()
            if not flush_ok:
                return sent_out
            sent_out += 1
        return sent_out

    async def post(self, agent_name: str):
        """
        Implementation of POST request handler for streaming chat API call.
        """

        metadata: Dict[str, Any] = self.get_metadata()
        service: AsyncAgentService = await self.get_service(agent_name, metadata)
        if service is None:
            return

        self.application.start_client_request(metadata, f"{agent_name}/streaming_chat")
        result_generator = None
        try:
            # Parse JSON body
            data = json.loads(self.request.body)
            result_generator = service.streaming_chat(data, metadata)

            # Set up headers for chunked response
            self.set_header("Content-Type", "application/json-lines")
            self.set_header("Transfer-Encoding", "chunked")
            # Flush headers immediately
            flush_ok: bool = await self.do_flush()
            if not flush_ok:
                # If we failed to flush our output,
                # most probably it's because connection is closed by a client.
                # Raise accordingly - we will handle this exception:
                raise tornado.iostream.StreamClosedError()

            async for result_dict in result_generator:
                result_str: str = json.dumps(result_dict) + "\n"
                self.write(result_str)
                flush_ok = await self.do_flush()
                if not flush_ok:
                    # Raise exception to be handled as a general
                    # "stream abruptly closed" case:
                    raise tornado.iostream.StreamClosedError()

        except (asyncio.CancelledError, tornado.iostream.StreamClosedError):
            # ensure generator is closed promptly
            # and re-raise as recommended
            if result_generator is not None:
                # Suppress possible exceptions: they are of no interest here.
                with contextlib.suppress(Exception):
                    await result_generator.aclose()
                    result_generator = None
            self.logger.info(metadata, "Request handler cancelled/stream closed.")
            raise

        except Exception as exc:  # pylint: disable=broad-exception-caught
            # Suppress possible exceptions: they are of no interest here.
            with contextlib.suppress(Exception):
                self.process_exception(exc)

        finally:
            # We are done with response stream:
            if result_generator is not None:
                with contextlib.suppress(Exception):
                    # It is possible we will call .aclose() twice
                    # on our result_generator - it is allowed and has no effect.
                    await result_generator.aclose()
            self.do_finish()
            self.application.finish_client_request(metadata, f"{agent_name}/streaming_chat", get_stats=True)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict


class ToolCall:
    """
    An interface representing a call to a tool
    """

    def get_id(self) -> str:
        """
        :return: The string id of this run
        """
        raise NotImplementedError

    def get_function_arguments(self) -> Dict[str, Any]:
        """
        :return: Returns a dictionary of the function arguments for the tool call
        """
        raise NotImplementedError

    def get_function_name(self) -> str:
        """
        :return: Returns the string name of the tool
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.service.utils.service_status import ServiceStatus


class ServerStatus:
    """
    Class for registering and reporting overall status of the server,
    primarily for interaction with external deployment environment.
    """

    def __init__(self, server_name: str):
        """
        Constructor.
        """
        self.server_name: str = server_name
        self.grpc_service: ServiceStatus = ServiceStatus("gRPC")
        self.http_service: ServiceStatus = ServiceStatus("http")
        self.updater: ServiceStatus = ServiceStatus("updater")

    def is_server_live(self) -> bool:
        """
        Return "live" status for the server
        """
        # If somebody calls this, we are at least alive
        return True

    def is_server_ready(self) -> bool:
        """
        Return "ready" status for the server
        """
        return \
            (not self.grpc_service.is_requested() or self.grpc_service.is_ready()) and \
            (not self.http_service.is_requested() or self.http_service.is_ready()) and \
            (not self.updater.is_requested() or self.updater.is_ready())

    def get_server_name(self) -> str:
        """
        Return server name
        """
        return self.server_name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

import os
from typing import Any, Dict

import json
from pyparsing.exceptions import ParseException
from pyparsing.exceptions import ParseSyntaxException

from leaf_common.persistence.easy.easy_hocon_persistence import EasyHoconPersistence
from leaf_common.persistence.interface.restorer import Restorer


class McpClientsInfoRestorer(Restorer):
    """
    Implementation of the Restorer interface that reads the MCP clients info file.
    NOTE: This class is highly experimental and implementation of MCP clients
    is very likely to change in future releases.
    """

    def restore(self, file_reference: str = None) -> Dict[str, Any]:
        """
        :param file_reference: The file reference to use when restoring.
                Default is None, implying the file reference is up to the
                implementation.
        :return: a dictionary with MCP clients information
        """
        file_path: str = file_reference
        if not file_path:
            file_path = os.environ.get("MCP_CLIENTS_INFO_FILE")
            if not file_path:
                # No clients info file specified.
                return None

        clients_info: Dict[str, Any] = None
        if file_path.endswith(".hocon"):
            hocon = EasyHoconPersistence()
            try:
                clients_info = hocon.restore(file_reference=file_path)
            except (ParseException, ParseSyntaxException) as exception:
                message: str = f"""
        There was an error parsing MCP clients info file "{file_path}".
        See the accompanying ParseException (above) for clues as to what might be
        syntactically incorrect in that file.
        """
                raise ParseException(message) from exception
        else:
            try:
                with open(file_path, "r", encoding="utf-8") as json_file:
                    clients_info = json.load(json_file)
            except FileNotFoundError:
                # Use the common verbiage below
                clients_info = None
            except json.decoder.JSONDecodeError as exception:
                message: str = f"""
        There was an error parsing MCP clients info file "{file_path}".
        See the accompanying JSONDecodeError exception (above) for clues as to what might be
        syntactically incorrect in that file.
        """
                raise ParseException(message) from exception
        if clients_info is None:
            message = f"Could not find MCP clients info file at path: {file_path}.\n" + """
            Some common problems include:
            * The file itself simply does not exist.
            * Path is not an absolute path and you are invoking the server from a place
              where the path is not reachable.
            * The path has a typo in it.

            Double-check the value of the MCP_CLIENTS_INFO_FILE env var and
            your current working directory (pwd).
            """
            raise FileNotFoundError(message)
        # Now, MCP endpoints urls could put in quotes, so strip them out.
        result_dict: Dict[str, Any] = {}
        for key, value in clients_info.items():
            use_key: str = key.replace(r'"', "")
            use_key = use_key.strip()
            result_dict[use_key] = value
        return result_dict




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
import http
from typing import Any
from typing import Dict
from typing import List

import json
import os
import asyncio

import tornado
from tornado.web import RequestHandler

from leaf_common.utils.async_atomic_counter import AsyncAtomicCounter
from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.generic.async_agent_service_provider import AsyncAgentServiceProvider
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.service.http.interfaces.agent_authorizer import AgentAuthorizer
from neuro_san.service.http.logging.http_logger import HttpLogger


class BaseRequestHandler(RequestHandler):
    """
    Abstract handler class for neuro-san API calls.
    Provides logic to inject neuro-san service specific data
    into local handler context.
    """
    request_id_counter: AsyncAtomicCounter = AsyncAtomicCounter()

    # pylint: disable=attribute-defined-outside-init
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def initialize(self,
                   agent_policy: AgentAuthorizer,
                   forwarded_request_metadata: List[str],
                   openapi_service_spec_path: str,
                   network_storage_dict: Dict[str, AgentNetworkStorage]):
        """
        This method is called by Tornado framework to allow
        injecting service-specific data into local handler context.
        :param agent_policy: abstract policy for agent requests
        :param forwarded_request_metadata: request metadata to forward.
        :param openapi_service_spec_path: file path to OpenAPI service spec.
        :param network_storage_dict: A dictionary of string (descripting scope) to
                    AgentNetworkStorage instance which keeps all the AgentNetwork instances
                    of a particular grouping.
        """

        self.agent_policy = agent_policy
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata
        self.openapi_service_spec_path: str = openapi_service_spec_path
        self.logger = HttpLogger(forwarded_request_metadata)
        self.network_storage_dict: Dict[str, AgentNetworkStorage] = network_storage_dict
        self.show_absent: bool = os.environ.get("SHOW_ABSENT_METADATA") is not None
        self.request_id: int = 0

        if os.environ.get("AGENT_ALLOW_CORS_HEADERS") is not None:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")
            headers: str = "Content-Type, Transfer-Encoding"
            metadata_headers: str = ", ".join(forwarded_request_metadata)
            if len(metadata_headers) > 0:
                headers += f", {metadata_headers}"
            # Set all allowed headers:
            self.set_header("Access-Control-Allow-Headers", headers)

    def get_metadata(self) -> Dict[str, Any]:
        """
        Extract user metadata defined by self.forwarded_request_metadata list
        from incoming request.
        :return: dictionary of user request metadata; possibly empty
        """
        metadata_dict: Dict[str, Any] =\
            BaseRequestHandler.get_request_metadata(
                self.request,
                self.forwarded_request_metadata,
                self.show_absent,
                self.logger)
        # Put in our own unique id so we have some way to track this request:
        user_request_id: str = metadata_dict.get("request_id", "None")
        if user_request_id == "None":
            metadata_dict["request_id"] = f"request-{self.request_id}"
        return metadata_dict

    @classmethod
    def get_request_metadata(cls, request,
                             forwarded_request_metadata: List[str],
                             show_absent: bool = False,
                             logger: HttpLogger = None
                             ) -> Dict[str, Any]:
        """
        Extract user metadata defined by forwarded_request_metadata list
        from incoming request.
        :param request: incoming http request
        :param forwarded_request_metadata: list of metadata keys
        :param show_absent: if True, will provide debug printout
               for all request metadata keys absent from incoming request headers;
               if False does nothing.
        :param logger: logger to use
        :return: dictionary of user request metadata; possibly empty
        """
        headers: Dict[str, Any] = request.headers
        result: Dict[str, Any] = {}
        for item_name in forwarded_request_metadata:
            if item_name in headers.keys():
                result[item_name] = headers[item_name]
            else:
                if show_absent and logger:
                    logger.warning({}, "MISSING METADATA VALUE: %s request %s", item_name, request.uri)
                result[item_name] = "None"
        return result

    async def get_service(self, agent_name: str, metadata: Dict[str, Any]) -> AsyncAgentService:
        """
        Get agent's AsyncAgentService for request execution
        :param agent_name: agent name
        :param metadata: metadata to be used for logging if necessary.
        :return: instance of AsyncAgentService if it is defined for this agent
                 None otherwise
        """
        service_provider: AsyncAgentServiceProvider = self.agent_policy.allow(agent_name)
        if service_provider is None:
            self.set_status(404)
            self.logger.error(metadata, "error: Invalid request path %s", self.request.path)
            self.do_finish()
            return None
        return service_provider.get_service()

    def process_exception(self, exc: Exception):
        """
        Process exception raised during request handling
        """
        if exc is None:
            return
        if isinstance(exc, json.JSONDecodeError):
            # Handle invalid JSON input
            self.set_status(400)
            self.write({"error": "Invalid JSON format"})
            self.logger.error(self.get_metadata(), "error: Invalid JSON format")
            return

        # General exception case:
        self.set_status(500)
        self.write({"error": "Internal server error"})
        self.logger.error(self.get_metadata(), "Internal server error: %s", str(exc))

    def data_received(self, chunk):
        """
        Method overrides abstract method of RequestHandler
        with no-op implementation.
        """
        return

    # Tornado can handle both syns and async versions of "prepare" method
    # pylint: disable=invalid-overridden-method
    async def prepare(self):
        if not self.application.is_serving():
            self.set_status(503)
            self.write({"error": "Server is shutting down"})
            self.logger.error(self.get_metadata(), "Server is shutting down")
            self.do_finish()
            return

        # Get unique request id in case we will need it:
        self.request_id = await self.request_id_counter.increment()

        self.logger.debug(self.get_metadata(), f"[REQUEST RECEIVED] {self.request.method} {self.request.uri}")

    def do_finish(self):
        """
        Wrapper for finish() call
        with check for closed client connection.
        """
        try:
            self.finish()
        except tornado.iostream.StreamClosedError:
            self.logger.warning(self.get_metadata(), "Finish: client closed connection unexpectedly.")

    async def do_flush(self) -> bool:
        """
        Wrapper for flush() call
        with check for closed client connection.
        """
        try:
            await self.flush()
            # What happens here: we have finished writing out one data item in our output stream,
            # and we have flushed Tornado output.
            # BUT: this does not guarantee in general that underlying TCP/IP transport
            # will flush its own buffers, so low-level buffering is still possible.
            # Result would be that several chat responses will be bunched together
            # and received by a client as one data piece.
            # If client is not ready for this, there will be problems.
            # SO: this real wall clock delay here helps to encourage underlying transport
            # to flush its own buffers - and we are good.
            # Duration of delay is speculative and maybe could be adjusted.
            # But best solution and reliable one: make client accept multiple data items
            # in one "get" request - as it should when dealing with streaming service.
            await asyncio.sleep(0.3)
            return True
        except tornado.iostream.StreamClosedError:
            self.logger.warning(self.get_metadata(), "Flush: client closed connection unexpectedly.")
            return False

    async def options(self, *_args, **_kwargs):
        """
        Handles OPTIONS requests for CORS support
        """
        # No body needed. Tornado will return a 204 No Content by default
        self.set_status(http.HTTPStatus.NO_CONTENT)
        self.do_finish()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT


class AgentSessionConstants:
    """
    Interface for shared constants between AgentSession and AsyncAgentSession
    """

    # Default gRPC port for the Agent Service
    DEFAULT_GRPC_PORT: int = 30011

    # Default port for the Agent HTTP Service
    # This port number will also be mentioned in its Dockerfile
    DEFAULT_HTTP_PORT: int = 8080

    DEFAULT_PORT: int = DEFAULT_HTTP_PORT




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from copy import deepcopy
from threading import Lock

from leaf_server_common.server.request_logger import RequestLogger

from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.service.generic.agent_service import AgentService
from neuro_san.service.generic.agent_server_logging import AgentServerLogging
from neuro_san.service.utils.server_context import ServerContext


# pylint: disable=too-many-instance-attributes
class AgentServiceProvider:
    """
    Class providing lazy construction of AgentService instance
    with given constructor parameters.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self,
                 request_logger: RequestLogger,
                 security_cfg: Dict[str, Any],
                 agent_name: str,
                 agent_network_provider: AgentNetworkProvider,
                 server_logging: AgentServerLogging,
                 server_context: ServerContext):
        """
        Constructor.
        :param request_logger: The instance of the RequestLogger that helps
                    keep track of stats
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  If None, uses insecure channel.
        :param agent_name: The agent name for the service
        :param agent_network_provider: The AgentNetworkProvider to use for the session.
        :param server_logging: An AgentServerLogging instance initialized so that
                        spawned asynchronous threads can also properly initialize
                        their logging.
        :param server_context: The ServerContext object with global-ish state
        """
        self.request_logger = request_logger
        self.security_cfg = deepcopy(security_cfg)
        self.server_logging: AgentServerLogging = server_logging
        self.agent_network_provider: AgentNetworkProvider = agent_network_provider
        self.agent_name: str = agent_name
        self.lock: Lock = Lock()
        self.server_context: ServerContext = server_context
        self.service_instance: AgentService = None

    def get_service(self) -> AgentService:
        """
        Get service instance.
        Create it if it is not instantiated yet.
        """
        if self.service_instance is None:
            with self.lock:
                if self.service_instance is None:
                    self.service_instance = AgentService(
                        self.request_logger,
                        self.security_cfg,
                        self.agent_name,
                        self.agent_network_provider,
                        self.server_logging,
                        self.server_context)
        return self.service_instance

    def service_created(self) -> bool:
        """
        Return True if service instance has already been instantiated;
               False otherwise
        """
        return self.service_instance is not None




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Union


from logging import getLogger
from logging import Logger

from neuro_san.interfaces.coded_tool import CodedTool


class Accountant(CodedTool):
    """
    A tool that updates a running cost each time it is called.
    """

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Union[Dict[str, Any], str]:
        """
        Updates the passed running cost each time it's called.
        :param args: A dictionary with the following keys:
                    "running_cost": the running cost to update.

        :param sly_data: A dictionary containing parameters that should be kept out of the chat stream.
                         Keys expected for this implementation are:
                         None

        :return: A dictionary containing:
                 "running_cost": the updated running cost.
        """
        tool_name = self.__class__.__name__
        logger: Logger = getLogger(self.__class__.__name__)

        logger.debug("========== Calling %s ==========", tool_name)
        # Parse the arguments
        logger.debug("args: %s", str(args))
        running_cost: float = float(args.get("running_cost"))

        # Increment the running cost not using value other 1
        # This would make a little hard if the LLM wanted to guess
        updated_running_cost: float = running_cost + 3.0

        tool_response = {
            "running_cost": updated_running_cost
        }
        logger.debug("-----------------------")
        logger.debug("%s response: %s", tool_name, tool_response)
        logger.debug("========== Done with %s ==========", tool_name)
        return tool_response




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.service.watcher.interfaces.storage_updater import StorageUpdater


class AbstractStorageUpdater(StorageUpdater):
    """
    Abstract base class for StorageUpdater implementations for common policy
    about checking for when it is needed to do update_storage().
    """

    def __init__(self, update_period_in_seconds: int):
        """
        Constructor

        :param update_period_in_seconds: An int describing how long this instance
                ideally wants to go between calls to update_storage().
        """
        self.last_update: float = 0.0
        self.update_period_in_seconds: int = update_period_in_seconds

    def start(self):
        """
        Perform start up.
        """
        raise NotImplementedError

    def update_storage(self):
        """
        Perform an update
        """
        raise NotImplementedError

    def get_update_period_in_seconds(self) -> int:
        """
        :return: An int describing how long this instance ideally wants to go between
                calls to update_storage().
        """
        return self.update_period_in_seconds

    def needs_updating(self, time_now_in_seconds: float) -> bool:
        """
        :param time_now_in_seconds: The current time in seconds.
                    We expect this to be from time.time()
        :return: True if this instance needs updating. False otherwise.
        """
        update_period: int = self.get_update_period_in_seconds()
        if update_period <= 0:
            # Never
            return False

        next_update: float = self.last_update + float(update_period)
        if time_now_in_seconds < next_update:
            # Not yet
            return False

        self.last_update = time_now_in_seconds
        return True




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Set

from os import environ

from asyncio import AbstractEventLoop
from asyncio import Event
from asyncio import run_coroutine_threadsafe
from logging import getLogger
from logging import Logger

from janus import Queue

from leaf_common.config.resolver_util import ResolverUtil

from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.interfaces.reservations_storage import ReservationsStorage
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.internals.reservations.agent_reservation import AgentReservation
from neuro_san.internals.reservations.abstract_agent_reservationist import AbstractAgentReservationist
from neuro_san.service.watcher.interfaces.abstract_storage_updater import AbstractStorageUpdater
from neuro_san.service.watcher.interfaces.startable import Startable


class TempNetworkStorageUpdater(AbstractStorageUpdater):
    """
    StorageUpdater implementation for temporary network updates.
    """

    def __init__(self, network_storage_dict: Dict[str, AgentNetworkStorage],
                 watcher_config: Dict[str, Any],
                 queues: Queue[AsyncCollatingQueue]):
        """
        Constructor

        :param network_storage_dict: A dictionary of string (descripting scope) to
                    AgentNetworkStorage instance which keeps all the AgentNetwork instances
                    of a particular grouping.
        :param watcher_config: A config dictionary for StorageUpdaters
        :param queues: A Queue of AsyncCollatingQueues for temp network deployment
        """
        super().__init__(watcher_config.get("temporary_network_update_period_seconds"))
        self.logger: Logger = getLogger(self.__class__.__name__)

        self.reservations_storage: Set[ReservationsStorage] = set()
        temp_storage: ReservationsStorage = network_storage_dict.get("temp")
        if temp_storage is not None:
            # If we don't have temp storage, we don't got nothin'
            self.reservations_storage.add(temp_storage)

            # Potentially create an external storage class
            storage_class_name: str = environ.get("AGENT_EXTERNAL_RESERVATIONS_STORAGE")
            external_storage: ReservationsStorage = ResolverUtil.create_instance(
                    storage_class_name,
                    "AGENT_EXTERNAL_RESERVATIONS_STORAGE env var",
                    ReservationsStorage)
            if external_storage is not None:
                self.reservations_storage.add(external_storage)

            # The ultimate target for sync is the temp storage
            for storage in self.reservations_storage:
                storage.set_sync_target(temp_storage)

        self.incoming: Queue[AsyncCollatingQueue] = queues
        self.queue_pool: Set[AsyncCollatingQueue] = set()
        self.reservationist = AbstractAgentReservationist(self.reservations_storage)

    def start(self):
        """
        Perform start up.
        """
        self.logger.info("Starting TempNetworkStorageUpdater with %d seconds period",
                         self.update_period_in_seconds)

        # Start any Startables
        for storage in self.reservations_storage:
            if isinstance(storage, Startable):
                storage.start()

    def update_storage(self):
        """
        Perform an update
        """
        # First sync any existing networks from potential external sources
        for storage in self.reservations_storage:
            storage.sync_reservations()

        # First expire any existing networks
        for storage in self.reservations_storage:
            storage.expire_reservations()

        # Get any new queues
        self.add_new_queues_to_pool()

        before = len(self.queue_pool)
        if before == 0:
            return

        # Process all our queues.
        # We take a copy because during processing we might remove a queue from the pool.
        self.logger.info("Updating temp storage from %d queues", before)
        for queue in self.queue_pool.copy():
            self.process_one_queue(queue)

        after = len(self.queue_pool)
        finished = before - after
        if finished > 0:
            self.logger.info("Temp storage from %d queues finished", finished)

    def add_new_queues_to_pool(self):
        """
        Checks our master queue of queues for any new additions
        and adds them to the queue pool we need to pay attention to.
        """
        while self.incoming.sync_q.qsize() > 0:
            async_collating_queue: AsyncCollatingQueue = self.incoming.sync_q.get()
            self.queue_pool.add(async_collating_queue)

    def process_one_queue(self, async_collating_queue: AsyncCollatingQueue):
        """
        Processes a single AsyncCollatingQueue from the pool

        :param async_collating_queue: The AsyncCollatingQueue to process
        """
        janus_queue: Queue = async_collating_queue.get_queue()

        # See what has come over this particular queue
        while janus_queue.sync_q.qsize() > 0:

            # Get an item off the queue
            queued_item: Dict[str, Any] = janus_queue.sync_q.get()
            if async_collating_queue.is_final_item(queued_item):
                # We have exhausted this queue.
                # No one needs to worry about it any more.
                self.queue_pool.remove(async_collating_queue)
                async_collating_queue.close()
                return

            self.process_one_queued_item(queued_item)

    def process_one_queued_item(self, queued_item: Dict[str, Any]):
        """
        Process a single item from one of the queues

        :param queued_item: A dictionary from a queue containing information
                as populated by AgentReservationist.deploy()
        """

        # Get the salient information from the dictionary that was queued.
        source: str = queued_item.get("source")
        deployment_dict: Dict[AgentReservation, Dict[str, Any]] = queued_item.get("deployment_dict")
        max_lifetime_in_seconds: float = queued_item.get("max_lifetime_in_seconds")

        # Do the deployment
        self.reservationist.deploy_together(deployment_dict, source, max_lifetime_in_seconds)

        # Maybe notify the deployer.
        event: Event = queued_item.get("event")
        if event is not None:
            event_loop: AbstractEventLoop = queued_item.get("event_loop")
            run_coroutine_threadsafe(self.reservationist.set_event(event), event_loop)

    def stop(self):
        """
        Perform stopping.
        """
        self.logger.info("Stopping TempNetworkStorageUpdater")

        # Stop any Startables
        for storage in self.reservations_storage:
            if isinstance(storage, Startable):
                storage.stop()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

import time

from neuro_san.interfaces.reservation import Reservation
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.reservations_storage import ReservationsStorage
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage


class ExpiringAgentNetworkStorage(AgentNetworkStorage, ReservationsStorage):
    """
    An AgentNetworkStorage instance where AgentNetworks are allowed to expire.
    """

    def __init__(self):
        """
        Constructor
        """
        super().__init__()
        self.reservations_table: Dict[str, Reservation] = {}
        self.last_modified: float = 0.0

    def set_sync_target(self, sync_target: ReservationsStorage):
        """
        :param sync_target: The ReservationsStorage where in-memory versions end up
        """
        # We don't need one of these

    def add_reservations(self, reservations_dict: Dict[Reservation, Dict[str, Any]],
                         source: str = None):
        """
        Add a set of reservations for agent networks en-masse

        :param reservations_dict: A mapping of Reservation -> agent network spec
        :param source: A string describing where the deployment was coming from
        """
        if not reservations_dict:
            # Nothing to do
            return

        # Figure out what's new vs what's not.
        # Need to do this while holding the lock
        added: List[str] = []
        replaced: List[str] = []
        with self.lock:
            for reservation, agent_spec in reservations_dict.items():

                agent_name: str = reservation.get_reservation_id()
                is_new = self.agents_table.get(agent_name) is None

                agent_network = AgentNetwork(agent_spec, agent_name)
                self.agents_table[agent_name] = agent_network
                self.reservations_table[agent_name] = reservation

                if is_new:
                    added.append(agent_name)
                else:
                    replaced.append(agent_name)

            self.last_modified = time.time()

        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            for agent_name in added:
                listener.agent_added(agent_name, self)
                self.logger.info("ADDED network for agent %s from %s", agent_name, source)
            for agent_name in replaced:
                listener.agent_modified(agent_name, self)
                self.logger.info("REPLACED network for agent %s from %s", agent_name, source)

    def sync_reservations(self):
        """
        Sync Reservations with some underlying data source
        """
        # Nothing to do here.  We are our own source of truth.

    def expire_reservations(self):
        """
        Remove Reservations that are expired
        """

        # First determine what has expired
        now: float = time.time()
        expired: List[str] = []

        for agent_name, reservation in self.reservations_table.items():
            if now > reservation.get_expiration_time_in_seconds():
                expired.append(agent_name)

        # Nothing to do?
        if len(expired) == 0:
            return

        # Do the dirty deeds.
        with self.lock:
            for agent_name in expired:
                self.reservations_table.pop(agent_name, None)
                self.agents_table.pop(agent_name, None)

            self.last_modified = time.time()

        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            for agent_name in expired:
                listener.agent_removed(agent_name, self)
                self.logger.info("REMOVED network for agent %s", agent_name)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

class ServiceStatus:
    """
    Class for registering and reporting overall status of the service,
    primarily for interaction with external deployment environment.
    """

    def __init__(self, service_name: str):
        """
        Constructor.
        """
        self.service_name: str = service_name
        self.service_requested: bool = True
        self.service_ready: bool = False

    def set_status(self, status: bool):
        """
        Set the status of a service
        """
        self.service_ready = status

    def is_ready(self) -> bool:
        """
        True if service is ready
        """
        return self.service_ready

    def set_requested(self, requested: bool):
        """
        Set if a service is requested by neuro-san server.
        """
        self.service_requested = requested

    def is_requested(self) -> bool:
        """
        True if service is requested.
        """
        return self.service_requested

    def get_service_name(self) -> str:
        """
        Return service name
        """
        return self.service_name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import List

import time

from threading import Lock
from threading import Thread

from tornado.web import Application
from tornado.web import ErrorHandler
from tornado.ioloop import IOLoop

from neuro_san.service.http.handlers.base_request_handler import BaseRequestHandler
from neuro_san.service.interfaces.event_loop_logger import EventLoopLogger


class HttpServerApp(Application):
    """
    Class provides customized Tornado application for neuro-san service -
    with redefined internal logger so we can include custom request metadata.
    """
    # pylint: disable=too-many-instance-attributes
    SHUTDOWN_TIMEOUT_SECONDS: int = 30

    def __init__(self, handlers,
                 requests_limit: int,
                 logger: EventLoopLogger,
                 forwarded_request_metadata: List[str]):
        """
        Constructor:
        :param handlers: list of request handlers
        :param requests_limit: limit for number of requests we can execute
        :param logger: logger to use
        :param forwarded_request_metadata: list of client metadata keys
        """
        # Call the base constructor
        super().__init__(handlers=handlers)
        self.total: int = 0
        self.num_processing: int = 0
        self.requests_stats: Dict[str, int] = {}
        self.requests_limit: int = requests_limit
        self.logger: EventLoopLogger = logger
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata
        self.serving: bool = True
        self.shutdown_initiated: bool = False
        self.lock: Lock = Lock()
        self.shutdown_thread = None

    def is_serving(self) -> bool:
        """
        Return True if server continues to serve requests,
        False otherwise.
        """
        return self.serving

    def start_client_request(self, metadata: Dict[str, Any], caller: str):
        """
        Register start of client request.
        :param metadata: request metadata
        :param caller: name of request client to be used for stats
        """
        self.logger.info(metadata, "Start %s", caller)
        with self.lock:
            self.num_processing += 1
            self.requests_stats[caller] = self.requests_stats.get(caller, 0) + 1

    def finish_client_request(self, metadata: Dict[str, Any],
                              caller: str, get_stats: bool = False):
        """
        Register finishing of client request.
        :param metadata: request metadata
        :param caller: name of request client to be used for stats
        :param get_stats: True if we need to log requests statistics,
                          False otherwise.
        """
        limit_reached: bool = False
        with self.lock:
            self.num_processing -= 1
            self.total += 1
            limit_reached = 0 <= self.requests_limit < self.total
        self.logger.info(metadata, "Finish %s", caller)
        if get_stats:
            self.logger.info(metadata, "Stats: %s", self.get_stats())
        # Now check if we reached requests limit:
        if limit_reached:
            self.serving = False
            self.initiate_shutdown()

    def do_shutdown(self, loop):
        """
        Poll for state with no executing requests
        or till we hit timeout:
        :param loop: event loop to stop
        """
        time_waited_seconds: int = 0
        wait_period_seconds = 2
        while time_waited_seconds < self.SHUTDOWN_TIMEOUT_SECONDS:
            if self.num_processing <= 0:
                break
            time.sleep(wait_period_seconds)
            time_waited_seconds += wait_period_seconds
        self.logger.info({}, "SERVER EXITING")
        self.stop_server(loop)

    def stop_server(self, loop):
        """
        Stop Tornado server event loop
        :param loop: event loop to stop
        """
        loop.add_callback(loop.stop)

    def initiate_shutdown(self):
        """
        Initiate server shutdown process
        """
        if self.shutdown_initiated:
            return
        self.shutdown_initiated = True
        self.logger.info({}, "Server request limit %d reached. Shutting down...", self.requests_limit)
        self.shutdown_thread = Thread(target=self.do_shutdown, args=(IOLoop.current(),), daemon=True)
        self.shutdown_thread.start()

    def get_stats(self) -> str:
        """
        Construct a string with current server requests statistics.
        """
        stats_dict: Dict[str, Any] = {
            "NumProcessing": self.num_processing,
            "Total": self.total
        }
        stats_dict.update(self.requests_stats)
        return str(stats_dict)

    def log_request(self, handler):
        if isinstance(handler, BaseRequestHandler):
            request = handler.request
            metadata: Dict[str, Any] = handler.get_metadata()
            status = handler.get_status()
            duration = 1000 * request.request_time()  # in milliseconds
            # handler.logger is our custom HttpLogger
            handler.logger.info(metadata, "%d %s %s (%s) %.2fms",
                                status, request.method, request.uri, request.remote_ip, duration)
        elif isinstance(handler, ErrorHandler):
            request = handler.request
            metadata: Dict[str, Any] =\
                BaseRequestHandler.get_request_metadata(request, self.forwarded_request_metadata)
            status = handler.get_status()
            duration = 1000 * request.request_time()  # in milliseconds
            # handler.logger is our custom HttpLogger
            self.logger.error(metadata, "%d %s %s (%s) %.2fms",
                              status, request.method, request.uri, request.remote_ip, duration)
        else:
            # Fall back to base request logger:
            super().log_request(handler)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Set

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.run_context.factory.master_toolbox_factory import MasterToolboxFactory
from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing


class ConnectivityReporter:
    """
    A class that knows how to report the connectivity of an entire AgentNetwork.

    Connectivity information comes as a series of AgentFramework
    messages, each of whose origin field reflects the name of the
    node and the content of the message is a JSON structure
    containing the list of tools that the node is connected to.

    *   The FrontMan is always sent first.
    *   Subsequent tool reporting proceeds in a breadth-first search, as per the
        ordering of the tools laid out in each agent spec.
    *   Cycles in the graph are mentioned in the tool reporting, but any node
        is only ever reported once.
    *   Hocon files can elect to hide the connectivity information from this reporting
        from any level on downstream by adding this block:
            "allow": {
                "connectivity": False
            }
        ...so if a network does not want connectivity reported at all, then this is only
        required in the front man's spec.
        Such a node will itself be reported, but with an empty tool list.
    *   External agents are reported but with an empty tool list.
        No effort to discover their internal connectivity is attempted.
        Maybe someday.
    """

    def __init__(self, inspector: AgentNetworkInspector):
        """
        Constructor

        :param inspector: The AgentNetworkInspector to use.
        """

        self.inspector: AgentNetworkInspector = inspector
        self.toolbox_factory: ContextTypeToolboxFactory = None

        if self.inspector is not None:
            config: Dict[str, Any] = self.inspector.get_config()
            self.toolbox_factory = MasterToolboxFactory.create_toolbox_factory(config)

    def report_network_connectivity(self) -> List[Dict[str, Any]]:
        """
        Share the connectivity information of the agent network in question
        :return: A list of connectivity information dictionaries each with the following keys:
            * origin  - The agent network node whose connectivity is being described
            * tools   - A list of tool nodes that are possible to reach from the origin

                        This might include references into external agent networks, perhaps hosted
                        on other servers.  Separate calls to those guys will need to be made
                        in order to gain information about their own connectivity, if this is
                        actually desired by the client.

                        Worth noting that server-side agent descriptions are allowed to
                        withhold connectivity info they deem private, or too much of an
                        implementation detail.  That is, connectivity reported is only
                        as much as the server wants a client to know.
        """
        # Load the toolbox factory once
        if self.toolbox_factory is not None:
            self.toolbox_factory.load()

        # Find the name of the front-man as a root node
        front_man: str = self.inspector.find_front_man()

        # Do a breadth-first traversal starting with the front-man
        reported_agents: Set[str] = set()
        connectivity: List[Dict[str, Any]] = self.report_node_connectivity(front_man, reported_agents)
        return connectivity

    def report_node_connectivity(self, agent_name: str, reported_agents: Set[str]) -> List[Dict[str, Any]]:
        """
        Share the connectivity information of a single node in the network.
        :param agent_name: The name of the agent spec dictionary to report on
        :param reported_agents: A list of agents that have been reported already.
                Prevents cycles.
        :return: A list of connectivity information dictionaries.
        """

        connectivity_list: List[Dict[str, Any]] = []
        agent_spec: Dict[str, Any] = None
        display_as: str = None
        metadata: Dict[str, Any] = None

        if not ExternalAgentParsing.is_external_agent(agent_name):

            # This is not an external agent, so get its spec to report on
            agent_spec: Dict[str, Any] = self.inspector.get_agent_tool_spec(agent_name)
            if agent_spec is None:
                # The agent referred to by the caller is not actually in the agent network.
                # As a hint, don't report anything, not even an empty tool list.
                return connectivity_list

            # Check to see if this node in the graph actually wants its connectivity
            # known to the outside world.
            extractor = DictionaryExtractor(agent_spec)
            allow_connectivity = bool(extractor.get("allow.connectivity", True))
            if not bool(allow_connectivity):

                # We are not allowing connectivity reporting from here on down
                agent_spec = None
        else:
            display_as = "external_agent"

        # Compile a tool list of what is referred to by the agent_spec.
        # For many reasons, this list could be empty.
        tool_list: List[str] = self.assemble_tool_list(agent_spec)

        if display_as is None:
            display_as = self.determine_display_as(agent_spec)

        if agent_spec is not None:
            metadata = agent_spec.get("metadata")

        connectivity_dict: Dict[str, Any] = {
            # Report the origin as the agent itself, so any client that receives
            "origin": agent_name,
            # Report the content of the tools list
            "tools": tool_list,
            # Report how the node wishes to be displayed
            "display_as": display_as
        }

        if metadata is not None:
            # Report any metadata about the node but only if we have it
            connectivity_dict["metadata"] = metadata

        # the message has the correct context about the tools listed in the content.
        reported_agents.add(agent_name)
        connectivity_list.append(connectivity_dict)

        # Recurse for a bread-first search.
        for tool in tool_list:
            # Don't report more than once for the same node to avoid cycles.
            if tool not in reported_agents:
                new_list: List[Dict[str, Any]] = self.report_node_connectivity(tool, reported_agents)
                if len(new_list) > 0:
                    connectivity_list.extend(new_list)

        return connectivity_list

    @staticmethod
    def assemble_tool_list(agent_spec: Dict[str, Any]) -> List[str]:
        """
        :param agent_spec: The agent spec to assemble a tool list from.
        :return: A list of tool names referred to by the agent spec.
        """

        # Keep a list of the tools to respect the ordering they were specified.
        tool_list: List[str] = []

        if agent_spec is None:
            # Nothing to report
            return tool_list

        extractor = DictionaryExtractor(agent_spec)

        # Keep a set of the combined sources of tools,
        # so connectivity only gets reported once.
        tool_set: Set[str] = set()

        # First check the tools of the run-of-the-mill agent spec
        empty_list: List[str] = []
        spec_tools: List[str] = extractor.get("tools", empty_list)

        # Add to our list in order, but without repeats
        for tool in spec_tools:
            if tool not in tool_set:
                tool_set.add(tool)
                tool_list.append(tool)

        # Next check a special case convention where a coded tool takes a dictionary of
        # key/value pairs mapping function -> tool name.
        empty_dict: Dict[str, Any] = {}
        args_tools = extractor.get("args.tools", empty_dict)

        if isinstance(args_tools, Dict):
            args_tools = args_tools.values()

        if isinstance(args_tools, List):
            # Add to our list in order, but without repeats
            for tool in args_tools:
                if tool not in tool_set:
                    tool_set.add(tool)
                    tool_list.append(tool)

        return tool_list

    def determine_display_as(self, agent_spec: Dict[str, Any]) -> str:
        """
        :param agent_spec: The agent spec to determine display_as from.
        :return: A string describing how the node wants to be seen.
        """
        # Attempt to get something node-specific from the spec.
        display_as: str = agent_spec.get("display_as")
        if display_as is not None:
            display_as = str(display_as)
            return display_as

        tool_name: str = agent_spec.get("toolbox")
        if tool_name is not None:
            # As a default, assume something from the toolbox is a lanchain_tool.
            display_as = "langchain_tool"

            tool_info: Dict[str, Any] = None
            if self.toolbox_factory is not None:
                tool_info = self.toolbox_factory.get_tool_info(tool_name)

            if tool_info is not None:
                if tool_info.get("display_as") is not None:
                    # Tool infos in a toolbox hocon can have their own display_as for potential branding
                    display_as = tool_info.get("display_as")
                elif tool_info.get("description") is not None:
                    # Shared coded tools need to at least specify their description,
                    # whereas langchain_tools get the description from their @tool annotation
                    display_as = "coded_tool"

        elif agent_spec.get("function") is not None:
            # If we have a function in the spec, the agent has arguments
            # it wants to be called with.
            if agent_spec.get("class") is not None:
                display_as = "coded_tool"
            else:
                display_as = "llm_agent"
        else:
            # Front-man
            display_as = "llm_agent"

        return display_as
